# Copyright 2018 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""Utilities for accessing the sqlite cache.

Note:
    All methods assume that nodes are integer-labeled.

"""
import sqlite3
import json
import os
import struct
import base64
import datetime

from dwave.system.cache.cache_manager import cache_file
from dwave.system.cache.schema import schema
from dwave.system.exceptions import MissingFluxBias


def cache_connect(database=None):
    """Returns a connection object to a sqlite database.

    Args:
        database (str, optional): The path to the database the user wishes
            to connect to. If not specified, a default is chosen using
            :func:`.cache_file`. If the special database name ':memory:'
            is given, then a temporary database is created in memory.

    Returns:
        :class:`sqlite3.Connection`

    """
    if database is None:
        database = cache_file()

    if os.path.isfile(database):
        # just connect to the database as-is
        conn = sqlite3.connect(database)
    else:
        # we need to populate the database
        conn = sqlite3.connect(database)
        conn.executescript(schema)

    with conn as cur:
        # turn on foreign keys, allows deletes to cascade.
        cur.execute("PRAGMA foreign_keys = ON;")

    conn.row_factory = sqlite3.Row

    return conn


def insert_chain(cur, chain, encoded_data=None):
    """Insert a chain into the cache.

    Args:
        cur (:class:`sqlite3.Cursor`):
            An sqlite3 cursor. This function is meant to be run within a :obj:`with` statement.

        chain (iterable):
            A collection of nodes. Chains in embedding act as one node.

        encoded_data (dict, optional):
            If a dictionary is provided, it will be populated with the serialized data. This is
            useful for preventing encoding the same information many times.

    Notes:
        This function assumes that the nodes in chain are index-labeled.

    """
    if encoded_data is None:
        encoded_data = {}

    if 'nodes' not in encoded_data:
        encoded_data['nodes'] = json.dumps(sorted(chain), separators=(',', ':'))
    if 'chain_length' not in encoded_data:
        encoded_data['chain_length'] = len(chain)

    insert = "INSERT OR IGNORE INTO chain(chain_length, nodes) VALUES (:chain_length, :nodes);"

    cur.execute(insert, encoded_data)


def iter_chain(cur):
    """Iterate over all of the chains in the database.

    Args:
        cur (:class:`sqlite3.Cursor`):
            An sqlite3 cursor. This function is meant to be run within a :obj:`with` statement.

    Yields:
        list: The chain.

    """
    select = "SELECT nodes FROM chain"
    for nodes, in cur.execute(select):
        yield json.loads(nodes)


def insert_system(cur, system_name, encoded_data=None):
    """Insert a system name into the cache.

    Args:
        cur (:class:`sqlite3.Cursor`):
            An sqlite3 cursor. This function is meant to be run within a :obj:`with` statement.

        system_name (str):
            The unique name of a system

        encoded_data (dict, optional):
            If a dictionary is provided, it will be populated with the serialized data. This is
            useful for preventing encoding the same information many times.

    """
    if encoded_data is None:
        encoded_data = {}

    if 'system_name' not in encoded_data:
        encoded_data['system_name'] = system_name

    insert = "INSERT OR IGNORE INTO system(system_name) VALUES (:system_name);"
    cur.execute(insert, encoded_data)


def iter_system(cur):
    """Iterate over all system names in the cache.

    Args:
        cur (:class:`sqlite3.Cursor`):
            An sqlite3 cursor. This function is meant to be run within a :obj:`with` statement.

    Yields:
        str: system name

    """
    select = "SELECT system_name FROM system"
    for system_name, in cur.execute(select):
        yield system_name


def insert_flux_bias(cur, chain, system, flux_bias, chain_strength, encoded_data=None):
    """Insert a flux bias offset into the cache.

    Args:
        cur (:class:`sqlite3.Cursor`):
            An sqlite3 cursor. This function is meant to be run within a :obj:`with` statement.

        chain (iterable):
            A collection of nodes. Chains in embedding act as one node.

        system (str):
            The unique name of a system.

        flux_bias (float):
            The flux bias offset associated with the given chain.

        chain_strength (float):
            The magnitude of the negative quadratic bias that induces the given chain in an Ising
            problem.

        encoded_data (dict, optional):
            If a dictionary is provided, it will be populated with the serialized data. This is
            useful for preventing encoding the same information many times.

    """
    if encoded_data is None:
        encoded_data = {}

    insert_chain(cur, chain, encoded_data)
    insert_system(cur, system, encoded_data)

    if 'flux_bias' not in encoded_data:
        encoded_data['flux_bias'] = _encode_real(flux_bias)
    if 'chain_strength' not in encoded_data:
        encoded_data['chain_strength'] = _encode_real(chain_strength)
    if 'insert_time' not in encoded_data:
        encoded_data['insert_time'] = datetime.datetime.now()

    insert = \
        """
        INSERT OR REPLACE INTO flux_bias(chain_id, system_id, insert_time, flux_bias, chain_strength)
        SELECT
            chain.id,
            system.id,
            :insert_time,
            :flux_bias,
            :chain_strength
        FROM chain, system
        WHERE
            chain.chain_length = :chain_length AND
            chain.nodes = :nodes AND
            system.system_name = :system_name;
        """

    cur.execute(insert, encoded_data)


def iter_flux_bias(cur):
    """Iterate over all flux biases in the cache.

    Args:
        cur (:class:`sqlite3.Cursor`):
            An sqlite3 cursor. This function is meant to be run within a :obj:`with` statement.

    Yields:
        tuple: A 4-tuple:

            list: The chain.

            str: The system name.

            float: The flux bias associated with the chain.

            float: The chain strength associated with the chain.

    """
    select = \
        """
        SELECT nodes, system_name, flux_bias, chain_strength FROM flux_bias_view;
        """

    for nodes, system, flux_bias, chain_strength in cur.execute(select):
        yield json.loads(nodes), system, _decode_real(flux_bias), _decode_real(chain_strength)


def _encode_real(v):
    """Encode real numbers as base 64 encoded little endian 8 byte floats."""
    bytes_ = struct.pack('<d', v)
    return base64.b64encode(bytes_)


def _decode_real(blob):
    """Inverse of _encode_real."""
    bytes_ = base64.b64decode(blob)
    return struct.unpack('<d', bytes_)[0]


def get_flux_biases_from_cache(cur, chains, system_name, chain_strength, max_age=3600):
    """Determine the flux biases for all of the the given chains, system and chain strength.

    Args:
        cur (:class:`sqlite3.Cursor`):
            An sqlite3 cursor. This function is meant to be run within a :obj:`with` statement.

        chains (iterable):
            An iterable of chains. Each chain is a collection of nodes. Chains in embedding act as
            one node.

        system_name (str):
            The unique name of a system.

        chain_strength (float):
            The magnitude of the negative quadratic bias that induces the given chain in an Ising
            problem.

        max_age (int, optional, default=3600):
            The maximum age (in seconds) for the flux_bias offsets.

    Returns:
        dict: A dict where the keys are the nodes in the chains and the values are the flux biases.

    """

    select = \
        """
        SELECT
            flux_bias
        FROM flux_bias_view WHERE
            chain_length = :chain_length AND
            nodes = :nodes AND
            chain_strength = :chain_strength AND
            system_name = :system_name AND
            insert_time >= :time_limit;
        """

    encoded_data = {'chain_strength': _encode_real(chain_strength),
                    'system_name': system_name,
                    'time_limit': datetime.datetime.now() + datetime.timedelta(seconds=-max_age)}

    flux_biases = {}
    for chain in chains:
        encoded_data['chain_length'] = len(chain)
        encoded_data['nodes'] = json.dumps(sorted(chain), separators=(',', ':'))

        row = cur.execute(select, encoded_data).fetchone()
        if row is None:
            raise MissingFluxBias
        flux_bias = _decode_real(*row)

        if flux_bias == 0:
            continue

        flux_biases.update({v: flux_bias for v in chain})

    return flux_biases


def insert_graph(cur, nodelist, edgelist, encoded_data=None):
    """Insert a graph into the cache.

    A graph is stored by number of nodes, number of edges and a
    json-encoded list of edges.

    Args:
        cur (:class:`sqlite3.Cursor`): An sqlite3 cursor. This function
            is meant to be run within a :obj:`with` statement.
        nodelist (list): The nodes in the graph.
        edgelist (list): The edges in the graph.
        encoded_data (dict, optional): If a dictionary is provided, it
            will be populated with the serialized data. This is useful for
            preventing encoding the same information many times.

    Notes:
        This function assumes that the nodes are index-labeled and range
        from 0 to num_nodes - 1.

        In order to minimize the total size of the cache, it is a good
        idea to sort the nodelist and edgelist before inserting.

    Examples:
        >>> nodelist = [0, 1, 2]
        >>> edgelist = [(0, 1), (1, 2)]
        >>> with pmc.cache_connect(':memory:') as cur:
        ...     pmc.insert_graph(cur, nodelist, edgelist)

        >>> nodelist = [0, 1, 2]
        >>> edgelist = [(0, 1), (1, 2)]
        >>> encoded_data = {}
        >>> with pmc.cache_connect(':memory:') as cur:
        ...     pmc.insert_graph(cur, nodelist, edgelist, encoded_data)
        >>> encoded_data['num_nodes']
        3
        >>> encoded_data['num_edges']
        2
        >>> encoded_data['edges']
        '[[0,1],[1,2]]'

    """
    if encoded_data is None:
        encoded_data = {}

    if 'num_nodes' not in encoded_data:
        encoded_data['num_nodes'] = len(nodelist)
    if 'num_edges' not in encoded_data:
        encoded_data['num_edges'] = len(edgelist)
    if 'edges' not in encoded_data:
        encoded_data['edges'] = json.dumps(edgelist, separators=(',', ':'))

    insert = \
        """
        INSERT OR IGNORE INTO graph(num_nodes, num_edges, edges)
        VALUES (:num_nodes, :num_edges, :edges);
        """

    cur.execute(insert, encoded_data)


def iter_graph(cur):
    """Iterate over all graphs in the cache.

    Args:
        cur (:class:`sqlite3.Cursor`): An sqlite3 cursor. This function
            is meant to be run within a :obj:`with` statement.

    Yields:
        tuple: A 2-tuple containing:

            list: The nodelist for a graph in the cache.

            list: the edgelist for a graph in the cache.

    Examples:
        >>> nodelist = [0, 1, 2]
        >>> edgelist = [(0, 1), (1, 2)]
        >>> with pmc.cache_connect(':memory:') as cur:
        ...     pmc.insert_graph(cur, nodelist, edgelist)
        ...     list(pmc.iter_graph(cur))
        [([0, 1, 2], [[0, 1], [1, 2]])]

    """
    select = """SELECT num_nodes, num_edges, edges from graph;"""
    for num_nodes, num_edges, edges in cur.execute(select):
        yield list(range(num_nodes)), json.loads(edges)


def insert_embedding(cur, source_nodelist, source_edgelist, target_nodelist, target_edgelist,
                     embedding, embedding_tag):
    """Insert an embedding into the cache.

    Args:
        cur (:class:`sqlite3.Cursor`):
            An sqlite3 cursor. This function is meant to be run within a :obj:`with` statement.

        source_nodelist (list):
            The nodes in the source graph. Should be integer valued.

        source_edgelist (list):
            The edges in the source graph.

        target_nodelist (list):
            The nodes in the target graph. Should be integer valued.

        target_edgelist (list):
            The edges in the target graph.

        embedding (dict):
            The mapping from the source graph to the target graph.
            Should be of the form {v: {s, ...}, ...} where v is a variable in the
            source model and s is a variable in the target model.

        embedding_tag (str):
            A string tag to associate with the embedding.

    """
    encoded_data = {}

    # first we need to encode the graphs and create the embedding id

    source_data = {}
    insert_graph(cur, source_nodelist, source_edgelist, source_data)
    encoded_data['source_edges'] = source_data['edges']
    encoded_data['source_num_nodes'] = source_data['num_nodes']
    encoded_data['source_num_edges'] = source_data['num_edges']

    target_data = {}
    insert_graph(cur, target_nodelist, target_edgelist, target_data)
    encoded_data['target_edges'] = target_data['edges']
    encoded_data['target_num_nodes'] = target_data['num_nodes']
    encoded_data['target_num_edges'] = target_data['num_edges']

    encoded_data['tag'] = embedding_tag

    insert_embedding = \
        """
        INSERT OR REPLACE INTO embedding(
            source_id,
            target_id,
            tag)
        SELECT
            source_graph.id,
            target_graph.id,
            :tag
        FROM
            graph 'source_graph',
            graph 'target_graph'
        WHERE
            source_graph.edges = :source_edges AND
            source_graph.num_nodes = :source_num_nodes AND
            source_graph.num_edges = :source_num_edges AND
            target_graph.edges = :target_edges AND
            target_graph.num_nodes = :target_num_nodes AND
            target_graph.num_edges = :target_num_edges
        """

    cur.execute(insert_embedding, encoded_data)

    # now each chain needs to be inserted

    insert_embedding_component = \
        """
        INSERT OR REPLACE INTO embedding_component(
            source_node,
            chain_id,
            embedding_id)
        SELECT
            :source_node,
            chain.id,
            embedding.id
        FROM
            graph 'source_graph',
            graph 'target_graph',
            chain,
            embedding
        WHERE
            source_graph.edges = :source_edges AND
            source_graph.num_nodes = :source_num_nodes AND
            target_graph.edges = :target_edges AND
            target_graph.num_nodes = :target_num_nodes AND
            embedding.source_id = source_graph.id AND
            embedding.target_id = target_graph.id AND
            embedding.tag = :tag AND
            chain.nodes = :nodes AND
            chain.chain_length = :chain_length
        """

    for v, chain in embedding.items():
        chain_data = {'source_node': v}
        insert_chain(cur, chain, chain_data)

        encoded_data.update(chain_data)

        cur.execute(insert_embedding_component, encoded_data)


def select_embedding_from_tag(cur, embedding_tag, target_nodelist, target_edgelist):
    """Select an embedding from the given tag and target graph.

    Args:
        cur (:class:`sqlite3.Cursor`):
            An sqlite3 cursor. This function is meant to be run within a :obj:`with` statement.

        source_nodelist (list):
            The nodes in the source graph. Should be integer valued.

        source_edgelist (list):
            The edges in the source graph.

        target_nodelist (list):
            The nodes in the target graph. Should be integer valued.

        target_edgelist (list):
            The edges in the target graph.

    Returns:
        dict: The mapping from the source graph to the target graph.
        In the form {v: {s, ...}, ...} where v is a variable in the
        source model and s is a variable in the target model.

    """
    encoded_data = {'num_nodes': len(target_nodelist),
                    'num_edges': len(target_edgelist),
                    'edges': json.dumps(target_edgelist, separators=(',', ':')),
                    'tag': embedding_tag}

    select = \
        """
        SELECT
            source_node,
            chain
        FROM
            embedding_component_view
        WHERE
            embedding_tag = :tag AND
            target_edges = :edges AND
            target_num_nodes = :num_nodes AND
            target_num_edges = :num_edges
        """

    embedding = {v: json.loads(chain) for v, chain in cur.execute(select, encoded_data)}
    return embedding


def select_embedding_from_source(cur, source_nodelist, source_edgelist,
                                 target_nodelist, target_edgelist):
    """Select an embedding from the source graph and target graph.

    Args:
        cur (:class:`sqlite3.Cursor`):
            An sqlite3 cursor. This function is meant to be run within a :obj:`with` statement.

        target_nodelist (list):
            The nodes in the target graph. Should be integer valued.

        target_edgelist (list):
            The edges in the target graph.

        embedding_tag (str):
            A string tag to associate with the embedding.

    Returns:
        dict: The mapping from the source graph to the target graph.
        In the form {v: {s, ...}, ...} where v is a variable in the
        source model and s is a variable in the target model.

    """
    encoded_data = {'target_num_nodes': len(target_nodelist),
                    'target_num_edges': len(target_edgelist),
                    'target_edges': json.dumps(target_edgelist, separators=(',', ':')),
                    'source_num_nodes': len(source_nodelist),
                    'source_num_edges': len(source_edgelist),
                    'source_edges': json.dumps(source_edgelist, separators=(',', ':'))}

    select = \
        """
        SELECT
            source_node,
            chain
        FROM
            embedding_component_view
        WHERE
            source_num_edges = :source_num_edges AND
            source_edges = :source_edges AND
            source_num_nodes = :source_num_nodes AND

            target_num_edges = :target_num_edges AND
            target_edges = :target_edges AND
            target_num_nodes = :target_num_nodes
        """

    embedding = {v: json.loads(chain) for v, chain in cur.execute(select, encoded_data)}
    return embedding



# Copyright 2018 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""
A :ref:`dimod composite <oceandocs:samplers_index>` that tiles small problems
multiple times to a structured sampler.

The :class:`.TilingComposite` class takes a problem that can fit on a small
:term:`Chimera` graph and replicates it across a larger Pegasus or
Chimera graph to obtain samples from multiple areas of the solver in one call.
For example, a 2x2 Chimera lattice could be tiled 64 times (8x8) on a
fully-yielded D-Wave 2000Q system (16x16).

See `Ocean Glossary <https://docs.ocean.dwavesys.com/en/stable/concepts/index.html>`_
for explanations of technical terms in descriptions of Ocean tools.

"""

from math import sqrt, ceil

import dimod
import dwave_networkx as dnx
import numpy as np

import dwave.embedding

__all__ = ['TilingComposite']


class TilingComposite(dimod.Sampler, dimod.Composite, dimod.Structured):
    """Composite to tile a small problem across a structured sampler.

    Enables parallel sampling on Chimera or Pegasus structured samplers of
    small problems. The small problem should be defined on a :term:`Chimera`
    graph of dimensions ``sub_m``, ``sub_n``, ``t``, or minor-embeddable to
    such a graph.

    Notation *CN* refers to a Chimera graph consisting of an NxN grid of unit
    cells, where each unit cell is a bipartite graph with shores of size t.
    The D-Wave 2000Q QPU supports a C16 Chimera graph: its 2048 qubits are
    logically mapped into a 16x16 matrix of unit cells of 8 qubits (t=4).
    See also the :func:`dwave_networkx.chimera_graph` function.

    Notation *PN* referes to a Pegasus graph consisting of a 3x(N-1)x(N-1) grid
    of cells, where each unit cell is a bipartite graph with shore of size t,
    supplemented with odd couplers (see ``nice_coordinate`` definition in
    the :func:`dwave_networkx.pegasus_graph` function). The
    Advantage QPU supports a P16 Pegasus graph: its qubits may be mapped to a
    3x15x15 matrix of unit cells, each of 8 qubits. This code supports tiling of
    Chimera-structured problems, with an option of additional odd-couplers,
    onto Pegasus. See also the :func:`dwave_networkx.pegasus_graph` function.

    A problem that can be minor-embedded in a single chimera unit cell, for
    example, can therefore be tiled across the unit cells of a D-Wave 2000Q as
    16x16 duplicates (or Advantage as 3x15x15 duplicates), subject to solver
    yield. This enables up to 256 (625) parallel samples per read.

    Args:
       sampler (:class:`dimod.Sampler`): Structured dimod sampler such as a
           :class:`~dwave.system.samplers.DWaveSampler()`.
       sub_m (int): Minimum number of Chimera unit cell rows required for
           minor-embedding a single instance of the problem.
       sub_n (int): Minimum number of Chimera unit cell columns required for
           minor-embedding a single instance of the problem.
       t (int, optional, default=4): Size of the shore within each Chimera unit
           cell.

    Examples:
       This example submits a two-variable QUBO problem representing a logical
       NOT gate to a D-Wave system. The QUBO---two nodes with biases of -1 that
       are coupled with strength 2---needs only any two coupled qubits and so is
       easily minor-embedded in a single unit cell.
       Composite :class:`.TilingComposite` tiles it multiple times for parallel solution:
       the two nodes should typically have opposite values.

       >>> from dwave.system import DWaveSampler, EmbeddingComposite
       >>> from dwave.system import TilingComposite
       ...
       >>> qpu_2000q = DWaveSampler(solver={'topology__type': 'chimera'})
       >>> sampler = EmbeddingComposite(TilingComposite(qpu_2000q, 1, 1, 4))
       >>> Q = {(1, 1): -1, (1, 2): 2, (2, 1): 0, (2, 2): -1}
       >>> sampleset = sampler.sample_qubo(Q)
       >>> len(sampleset)> 1
       True

    See `Ocean Glossary <https://docs.ocean.dwavesys.com/en/stable/concepts/index.html>`_
    for explanations of technical terms in descriptions of Ocean tools.

    """
    nodelist = None
    """list: List of active qubits for the structured solver."""

    edgelist = None
    """list: List of active couplers for the structured solver."""

    parameters = None
    """dict[str, list]: Parameters in the form of a dict."""

    properties = None
    """dict: Properties in the form of a dict."""

    children = None
    """list: The single wrapped structured sampler."""

    embeddings = []
    """list: Embeddings into each available tile on the structured solver."""

    def __init__(self, sampler, sub_m, sub_n, t=4):

        self.parameters = sampler.parameters.copy()
        self.properties = properties = {'child_properties': sampler.properties}

        tile = dnx.chimera_graph(sub_m, sub_n, t)
        self.nodelist = sorted(tile.nodes)
        self.edgelist = sorted(sorted(edge) for edge in tile.edges)
        # dimod.Structured abstract base class automatically populates adjacency
        # and structure as mixins based on nodelist and edgelist

        if not isinstance(sampler, dimod.Structured):
            # we could also just tile onto the unstructured sampler but in that
            # case we would need to know how many tiles to use
            raise ValueError("given child sampler should be structured")
        self.children = [sampler]
        # Chimera values (unless pegasus specified)
        num_sublattices=1
        nodes_per_cell = t * 2
        edges_per_cell = t * t
        if not ('topology' in sampler.properties
            and 'type' in sampler.properties['topology']
            and 'shape' in sampler.properties['topology']):
            raise ValueError('To use this composite it is necessary for the'
                             'structured sampler to have an explicit topology'
                             '(sampler.properties[\'topology\']). Necessary'
                             'fields are \'type\' and \'shape\'. ')
        if sampler.properties['topology']['type'] == 'chimera':
            if len(sampler.properties['topology']['shape']) != 3:
                raise ValueError('topology shape is not of length 3 '
                                 '(not compatible with chimera)')
            if sampler.properties['topology']['shape'][2] != t:
                raise ValueError('Tiling methodology requires that solver'
                                 'and subproblem have identical shore size')
            m = sampler.properties['topology']['shape'][0]
            n = sampler.properties['topology']['shape'][1]
        else:
            if len(sampler.properties['topology']['shape']) != 1:
                raise ValueError('topology shape is not of length 1 '
                                 '(not compatible with pegasus)')
            # Full yield in odd-couplers also required.
            # Generalizes chimera subgraph requirement and leads to some
            # simplification of expressions, but at with a cost in cell-yield
            edges_per_cell += t
            # Square solvers only by pegasus lattice definition PN yields
            # 3 by N-1 by N-1 cells:
            num_sublattices=3
            m = n = sampler.properties['topology']['shape'][0] - 1
            if t!=4:
                raise ValueError(
                    't=4 for all pegasus processors, value is not typically'
                    'stored in solver properties and is difficult to infer.'
                    'Therefore only the value t=4 is supported.')


        if num_sublattices==1:
            # Chimera defaults. Appended coordinates (treat as first and only sublattice)
            system = dnx.chimera_graph(m, n, t,
                                       node_list=sampler.structure.nodelist,
                                       edge_list=sampler.structure.edgelist)

            c2i = {(0, *chimera_index) : linear_index
                   for (linear_index, chimera_index)
                   in system.nodes(data='chimera_index')}
        else:
            system = dnx.pegasus_graph(m,
                                       node_list=sampler.structure.nodelist,
                                       edge_list=sampler.structure.edgelist)
            # Vector specification in terms of nice coordinates:
            c2i = {dnx.pegasus_coordinates(m+1).linear_to_nice(linear_index):
                   linear_index for linear_index in system.nodes()}

        sub_c2i = {chimera_index: linear_index for (linear_index, chimera_index)
                   in tile.nodes(data='chimera_index')}

        # Count the connections between these qubits
        def _between(qubits1, qubits2):
            edges = [edge for edge in system.edges if edge[0] in qubits1
                     and edge[1] in qubits2]
            return len(edges)

        # Get the list of qubits in a cell
        def _cell_qubits(s, i, j):
            return [c2i[(s, i, j, u, k)] for u in range(2) for k in range(t)
                    if (s, i, j, u, k) in c2i]

        # get a mask of complete cells
        cells = [[[False for _ in range(n)] for _ in range(m)]
                 for _ in range(num_sublattices)]

        for s in range(num_sublattices):
            for i in range(m):
                for j in range(n):
                    qubits = _cell_qubits(s, i, j)
                    cells[s][i][j] = (
                        len(qubits) == nodes_per_cell
                        and _between(qubits, qubits) == edges_per_cell)

        # List of 'embeddings'
        self.embeddings = properties['embeddings'] = embeddings = []

        # For each possible chimera cell check if the next few cells are complete
        for s in range(num_sublattices):
            for i in range(m + 1 - sub_m):
                for j in range(n + 1 - sub_n):

                    # Check if the sub cells are matched
                    match = all(cells[s][i + sub_i][j + sub_j]
                                for sub_i in range(sub_m)
                                for sub_j in range(sub_n))

                    # Check if there are connections between the cells.
                    # Both Pegasus and Chimera have t vertical and t horizontal between cells:
                    for sub_i in range(sub_m):
                        for sub_j in range(sub_n):
                            if sub_m > 1 and sub_i < sub_m - 1:
                                match &= _between(
                                    _cell_qubits(s, i + sub_i, j + sub_j),
                                    _cell_qubits(s, i + sub_i + 1, j + sub_j)) == t
                            if sub_n > 1 and sub_j < sub_n - 1:
                                match &= _between(
                                    _cell_qubits(s, i + sub_i, j + sub_j),
                                    _cell_qubits(s, i + sub_i, j + sub_j + 1)) == t

                    if match:
                        # Pull those cells out into an embedding.
                        embedding = {}
                        for sub_i in range(sub_m):
                            for sub_j in range(sub_n):
                                cells[s][i + sub_i][j + sub_j] = False  # Mark cell as matched
                                for u in range(2):
                                    for k in range(t):
                                        embedding[sub_c2i[sub_i, sub_j, u, k]] = {
                                            c2i[(s,i + sub_i, j + sub_j, u, k)]}

                        self.embeddings.append(embedding)

        if len(self.embeddings) == 0:
            raise ValueError("no tile embeddings found; "
                             "is the sampler Pegasus or Chimera structured?")

    @dimod.bqm_structured
    def sample(self, bqm, **kwargs):
        """Sample from the specified binary quadratic model.

        Args:
            bqm (:obj:`dimod.BinaryQuadraticModel`):
                Binary quadratic model to be sampled from.

            **kwargs:
                Optional keyword arguments for the sampling method, specified per solver.

        Returns:
            :class:`dimod.SampleSet`

        Examples:
            This example submits a simple Ising problem of just two variables on a
            D-Wave system.
            Because the problem fits in a single :term:`Chimera` unit cell, it is tiled
            across the solver's entire Chimera graph, resulting in multiple samples
            (the exact number depends on the working Chimera graph of the D-Wave system).

            >>> from dwave.system import DWaveSampler, EmbeddingComposite
            >>> from dwave.system import TilingComposite
            ...
            >>> qpu_2000q = DWaveSampler(solver={'topology__type': 'chimera'})
            >>> sampler = EmbeddingComposite(TilingComposite(qpu_2000q, 1, 1, 4))
            >>> response = sampler.sample_ising({},{('a', 'b'): 1})
            >>> len(response)    # doctest: +SKIP
            246

        See `Ocean Glossary <https://docs.ocean.dwavesys.com/en/stable/concepts/index.html>`_
        for explanations of technical terms in descriptions of Ocean tools.

        """

        # apply the embeddings to the given problem to tile it across the child sampler
        embedded_bqm = dimod.BinaryQuadraticModel.empty(bqm.vartype)
        __, __, target_adjacency = self.child.structure
        for embedding in self.embeddings:
            embedded_bqm.update(dwave.embedding.embed_bqm(bqm, embedding, target_adjacency))

        # solve the problem on the child system
        tiled_response = self.child.sample(embedded_bqm, **kwargs)

        responses = []

        for embedding in self.embeddings:
            embedding = {v: chain for v, chain in embedding.items() if v in bqm.variables}

            responses.append(dwave.embedding.unembed_sampleset(tiled_response, embedding, bqm))

        answer = dimod.concatenate(responses)
        answer.info.update(tiled_response.info)

        return answer

    @property
    def num_tiles(self):
        """Number of tiles available for replicating the problem."""
        return len(self.embeddings)



# Copyright 2020 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import itertools
import os
import unittest
import warnings

import dimod

from dwave.cloud.exceptions import ConfigFileError, SolverNotFoundError
from dwave.system import DWaveCliqueSampler

from parameterized import parameterized

_SAMPLERS = {}


def get_sampler(topology):
    if topology in _SAMPLERS:
        return _SAMPLERS[topology]
    try:
        _SAMPLERS[topology] = DWaveCliqueSampler(solver=dict(topology__type=topology.lower()))
        return _SAMPLERS[topology]
    except (ValueError, ConfigFileError, SolverNotFoundError):
        raise unittest.SkipTest(f"no {topology}-structured QPU available")


@unittest.skipIf(os.getenv('SKIP_INT_TESTS'), "Skipping integration test.")
class TestDWaveCliqueSampler(unittest.TestCase):
    @parameterized.expand([['Chimera'], ['Pegasus'], ['Zephyr']])
    def test_maximum_ferromagnet(self, topology):
        sampler = get_sampler(topology)

        dimod.testing.assert_sampler_api(sampler)

        bqm = dimod.BinaryQuadraticModel('SPIN')
        for u, v in itertools.combinations(sampler.largest_clique(), 2):
            bqm.quadratic[u, v] = -1

        sampler.sample(bqm).resolve()

    @parameterized.expand(itertools.product(('Pegasus', 'Zephyr'), (None, 0.5)))
    def test_per_qubit_coupling_range(self, topology, chain_strength):
        sampler = get_sampler(topology)
        n = sampler.largest_clique_size

        bqm = dimod.BinaryQuadraticModel({},
                {(u, v): -2 for u in range(n) for v in range(u+1, n)}, 'SPIN')

        with warnings.catch_warnings(record=True) as w:
            sampler.sample(bqm, chain_strength=chain_strength).resolve()

        if topology == 'Pegasus':
            limit_name = 'per_qubit_coupling_range'
        else:
            limit_name = 'per_group_coupling_range'

        if chain_strength is not None:
            self.assertEqual(len(w), 1)
            self.assertIn(f'{limit_name} is violated', str(w[0].message))
        else:
            self.assertEqual(w, [])



# Copyright 2018 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import os
import unittest
import warnings

import dimod

from dwave.cloud.exceptions import ConfigFileError, SolverNotFoundError

from dwave.system import DWaveSampler, EmbeddingComposite


@unittest.skipIf(os.getenv('SKIP_INT_TESTS'), "Skipping integration test.")
class TestEmbeddingCompositeExactSolver(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        try:
            cls.qpu = DWaveSampler(
                solver=dict(initial_state=True, anneal_schedule=True))
        except (ValueError, ConfigFileError, SolverNotFoundError):
            raise unittest.SkipTest("no qpu available")

    @classmethod
    def tearDownClass(cls):
        cls.qpu.client.close()

    def test_initial_state(self):
        sampler = EmbeddingComposite(self.qpu)

        bqm = dimod.BinaryQuadraticModel.from_ising({'a': 2.0, 'b': -2.0},
                                                    {('a', 'b'): -1})

        kwargs = {'initial_state': {'a': 1, 'b': 1},
                  'anneal_schedule': [(0, 1), (55.0, 0.45),
                                      (155.0, 0.45), (210.0, 1)]}

        sampler.sample(bqm, **kwargs).resolve()

    def test_many_bqm_async(self):
        sampler = EmbeddingComposite(self.qpu)

        # in the future it would be good to test a wide variety of BQMs,
        # see https://github.com/dwavesystems/dimod/issues/671
        # but for now let's just test a few
        bqm0 = dimod.BinaryQuadraticModel.from_ising({'a': 2.0, 'b': -2.0},
                                                     {('a', 'b'): -1})
        bqm1 = dimod.BinaryQuadraticModel.from_ising({2: 4},
                                                     {(0, 1): 1.5, (1, 2): 5})

        samplesets0 = []
        samplesets1 = []
        for _ in range(10):
            # this should be async
            samplesets0.append(sampler.sample(bqm0))
            samplesets1.append(sampler.sample(bqm1))

        if all(ss.done() for ss in samplesets0):
            warnings.warn("Sampler calls appear to be synchronous")

        for ss0, ss1 in zip(samplesets0, samplesets1):
            dimod.testing.assert_sampleset_energies(ss0, bqm0)
            dimod.testing.assert_sampleset_energies(ss1, bqm1)

        self.assertTrue(all(ss.done() for ss in samplesets0))
        self.assertTrue(all(ss.done() for ss in samplesets1))



# Copyright 2018 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
# ================================================================================================
import unittest

import networkx as nx

from dwave.embedding.exceptions import MissingChainError, ChainOverlapError, DisconnectedChainError
from dwave.embedding.exceptions import InvalidNodeError, MissingEdgeError
from dwave.embedding.diagnostic import diagnose_embedding, is_valid_embedding, verify_embedding


class TestEmbeddingDiagnostic(unittest.TestCase):
    def test_missing_chain(self):
        k2 = nx.complete_graph(2)
        emb = {0: [0]}

        self.assertRaises(MissingChainError, lambda: verify_embedding(emb, k2, k2))
        self.assertFalse(is_valid_embedding(emb, k2, k2))

        etypes = set()
        for err in diagnose_embedding(emb, k2, k2):
            etypes.add(err[0])
        self.assertEqual(etypes, {MissingChainError})

    def test_chain_overlap(self):
        k2 = nx.complete_graph(2)
        emb = {0: [0], 1: [0, 1]}

        self.assertRaises(ChainOverlapError, lambda: verify_embedding(emb, k2, k2))
        self.assertFalse(is_valid_embedding(emb, k2, k2))

        etypes = set()
        for err in diagnose_embedding(emb, k2, k2):
            etypes.add(err[0])
        self.assertEqual(etypes, {ChainOverlapError})

    def test_chain_overlap_with_edges(self):
        #this is made for compatibility with minorminer; to verify that "overlapped
        #embeddings" don't report spurious MissingEdgeErrors
        k5 = nx.complete_graph(5)
        k4 = nx.complete_graph(4)
        emb = {i:[i%4, (i+1)%4] for i in k5}

        self.assertRaises(ChainOverlapError, lambda: verify_embedding(emb, k5, k4))
        self.assertFalse(is_valid_embedding(emb, k5, k4))

        etypes = set()
        for err in diagnose_embedding(emb, k5, k4):
            etypes.add(err[0])
        self.assertEqual(etypes, {ChainOverlapError})

    def test_chain_disconnect(self):
        k2 = nx.complete_graph(2)
        p3 = nx.path_graph(3)
        emb = {0: [1], 1: [0, 2]}

        self.assertRaises(DisconnectedChainError, lambda: verify_embedding(emb, k2, p3))
        self.assertFalse(is_valid_embedding(emb, k2, p3))

        etypes = set()
        for err in diagnose_embedding(emb, k2, p3):
            etypes.add(err[0])
        self.assertEqual(etypes, {DisconnectedChainError})

    def test_invalid_node(self):
        k2 = nx.complete_graph(2)
        emb = {0: [0], 1: [2]}

        self.assertRaises(InvalidNodeError, lambda: verify_embedding(emb, k2, k2))
        self.assertFalse(is_valid_embedding(emb, k2, k2))

        etypes = set()
        for err in diagnose_embedding(emb, k2, k2):
            etypes.add(err[0])
        self.assertEqual(etypes, {InvalidNodeError})

    def test_missing_edge(self):
        k2 = nx.complete_graph(2)
        e2 = nx.empty_graph(2)
        emb = {0: [0], 1: [1]}

        self.assertRaises(MissingEdgeError, lambda: verify_embedding(emb, k2, e2))
        self.assertFalse(is_valid_embedding(emb, k2, e2))

        etypes = set()
        for err in diagnose_embedding(emb, k2, e2):
            etypes.add(err[0])
        self.assertEqual(etypes, {MissingEdgeError})

    def test_valid(self):
        k2 = nx.complete_graph(2)
        emb = {0: [0], 1: [1]}

        verify_embedding(emb, k2, k2)
        self.assertTrue(is_valid_embedding(emb, k2, k2))

        for err in diagnose_embedding(emb, k2, k2):
            raise err[0](*err[1:])



# Copyright 2020 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""
A :std:doc:`dimod sampler <oceandocs:docs_dimod/reference/samplers>` for Leap's hybrid solvers.
"""

from typing import Any, Dict, List, Optional

import numpy as np
from warnings import warn
from numbers import Number
from collections import abc

import dimod

try:
    # dimod 0.10.x
    from dimod.binary.binary_quadratic_model import BQM

    bqm_to_file = BQM.to_file
except ImportError:
    # dimod 0.9.x
    from dimod import AdjVectorBQM as BQM
    from dimod.serialization.fileview import FileView

    bqm_to_file = FileView

from dwave.cloud import Client
from dwave.system.utilities import classproperty, FeatureFlags


__all__ = ['LeapHybridSampler',
           'LeapHybridBQMSampler',
           'LeapHybridDQMSampler',
           'LeapHybridCQMSampler',
           ]


class LeapHybridSampler(dimod.Sampler):
    """A class for using Leap's cloud-based hybrid BQM solvers.

    Leap’s quantum-classical hybrid BQM solvers are intended to solve arbitrary
    application problems formulated as binary quadratic models (BQM).

    You can configure your :term:`solver` selection and usage by setting parameters,
    hierarchically, in a configuration file, as environment variables, or
    explicitly as input arguments, as described in
    `D-Wave Cloud Client <https://docs.ocean.dwavesys.com/en/stable/docs_cloud/sdk_index.html>`_.

    :ref:`dwave-cloud-client <sdk_index_cloud>`'s
    :meth:`~dwave.cloud.client.Client.get_solvers` method filters solvers you have
    access to by `solver properties <https://docs.dwavesys.com/docs/latest/c_solver_properties.html>`_
    ``category=hybrid`` and ``supported_problem_type=bqm``. By default, online
    hybrid BQM solvers are returned ordered by latest ``version``.

    The default specification for filtering and ordering solvers by features is
    available as :attr:`.default_solver` property. Explicitly specifying a
    solver in a configuration file, an environment variable, or keyword
    arguments overrides this specification. See the example below on how to
    extend it instead.

    Args:
        **config:
            Keyword arguments passed to :meth:`dwave.cloud.client.Client.from_config`.

    Examples:
        This example builds a random sparse graph and uses a hybrid solver to find a
        maximum independent set.

        >>> import dimod
        >>> import networkx as nx
        >>> import dwave_networkx as dnx
        >>> import numpy as np
        >>> from dwave.system import LeapHybridSampler
        ...
        >>> # Create a maximum-independent set problem from a random graph
        >>> problem_node_count = 300
        >>> G = nx.random_geometric_graph(problem_node_count, radius=0.0005*problem_node_count)
        >>> qubo = dnx.algorithms.independent_set.maximum_weighted_independent_set_qubo(G)
        >>> bqm = dimod.BQM.from_qubo(qubo)
        ...
        >>> # Find a good solution
        >>> sampler = LeapHybridSampler()       # doctest: +SKIP
        >>> sampleset = sampler.sample(bqm)     # doctest: +SKIP

        This example specializes the default solver selection by filtering out
        bulk BQM solvers. (Bulk solvers are throughput-optimal for heavy/batch
        workloads, have a higher start-up latency, and are not well suited for
        live workloads. Not all Leap accounts have access to bulk solvers.)

        >>> from dwave.system import LeapHybridSampler
        ...
        >>> solver = LeapHybridSampler.default_solver
        >>> solver.update(name__regex=".*(?<!bulk)$")       # name shouldn't end with "bulk"
        >>> sampler = LeapHybridSampler(solver=solver)      # doctest: +SKIP
        >>> sampler.solver        # doctest: +SKIP
        BQMSolver(id='hybrid_binary_quadratic_model_version2')

    """

    _INTEGER_BQM_SIZE_THRESHOLD = 10000

    @classproperty
    def default_solver(cls):
        """dict: Features used to select the latest accessible hybrid BQM solver."""
        return dict(supported_problem_types__contains='bqm',
                    order_by='-properties.version')

    def __init__(self, **config):
        # strongly prefer hybrid solvers; requires kwarg-level override
        config.setdefault('client', 'hybrid')

        # default to short-lived session to prevent resets on slow uploads
        config.setdefault('connection_close', True)

        if FeatureFlags.hss_solver_config_override:
            # use legacy behavior (override solver config from env/file)
            solver = config.setdefault('solver', {})
            if isinstance(solver, abc.Mapping):
                solver.update(self.default_solver)

        # prefer the latest hybrid BQM solver available, but allow for an easy
        # override on any config level above the defaults (file/env/kwarg)
        defaults = config.setdefault('defaults', {})
        if not isinstance(defaults, abc.Mapping):
            raise TypeError("mapping expected for 'defaults'")
        defaults.update(solver=self.default_solver)

        self.client = Client.from_config(**config)
        self.solver = self.client.get_solver()

        # check user-specified solver conforms to our requirements
        if self.properties.get('category') != 'hybrid':
            raise ValueError("selected solver is not a hybrid solver.")
        if 'bqm' not in self.solver.supported_problem_types:
            raise ValueError("selected solver does not support the 'bqm' problem type.")

    @property
    def properties(self) -> Dict[str, Any]:
        """Solver properties as returned by a SAPI query.

        `Solver properties <https://docs.dwavesys.com/docs/latest/c_solver_properties.html>`_
        are dependent on the selected solver and subject to change.
        """
        try:
            return self._properties
        except AttributeError:
            self._properties = properties = self.solver.properties.copy()
            return properties

    @property
    def parameters(self) -> Dict[str, list]:
        """Solver parameters in the form of a dict, where keys are
        keyword parameters accepted by a SAPI query and values are lists of properties in
        :attr:`~dwave.system.samplers.LeapHybridSampler.properties` for each key.

        `Solver parameters <https://docs.dwavesys.com/docs/latest/c_solver_parameters.html>`_
        are dependent on the selected solver and subject to change.
        """
        try:
            return self._parameters
        except AttributeError:
            parameters = {param: ['parameters']
                          for param in self.properties['parameters']}
            parameters.update(label=[])
            self._parameters = parameters
            return parameters

    def sample(self, bqm, time_limit=None, **kwargs):
        """Sample from the specified binary quadratic model.

        Args:
            bqm (:obj:`dimod.BinaryQuadraticModel`):
                Binary quadratic model.

            time_limit (int):
                Maximum run time, in seconds, to allow the solver to work on the
                problem. Must be at least the minimum required for the number of
                problem variables, which is calculated and set by default.

                :meth:`~dwave.system.samplers.LeapHybridSampler.min_time_limit`
                calculates (and describes) the minimum time for your problem.

            **kwargs:
                Optional keyword arguments for the solver, specified in
                :attr:`~dwave.system.samplers.LeapHybridSampler.parameters`.

        Returns:
            :class:`~dimod.SampleSet`: Sample set constructed from a (non-blocking)
            :class:`~concurrent.futures.Future`-like object.

        Examples:
            This example builds a random sparse graph and uses a hybrid solver to
            find a maximum independent set.

            >>> import dimod
            >>> import networkx as nx
            >>> import dwave_networkx as dnx
            >>> import numpy as np
            ...
            >>> # Create a maximum-independent set problem from a random graph
            >>> problem_node_count = 300
            >>> G = nx.random_geometric_graph(problem_node_count, radius=0.0005*problem_node_count)
            >>> qubo = dnx.algorithms.independent_set.maximum_weighted_independent_set_qubo(G)
            >>> bqm = dimod.BQM.from_qubo(qubo)
            ...
            >>> # Find a good solution
            >>> sampler = LeapHybridSampler()    # doctest: +SKIP
            >>> sampleset = sampler.sample(bqm)           # doctest: +SKIP

        """
        if not isinstance(bqm, BQM):
            bqm = BQM(bqm)

        num_vars = bqm.num_variables

        if time_limit is None:
            time_limit = self.min_time_limit(bqm)
        if not isinstance(time_limit, Number):
            raise TypeError("time limit must be a number")
        if time_limit < self.min_time_limit(bqm):
            msg = ("time limit for problem size {} must be at least {}"
                   ).format(num_vars, self.min_time_limit(bqm))
            raise ValueError(msg)

        # for very large BQMs, it is better to send the unlabelled version,
        # to save on serializating the labels in both directions.
        # Note that different hybrid solvers accept different numbers of
        # variables and they might be lower than this threshold
        if num_vars > self._INTEGER_BQM_SIZE_THRESHOLD:
            return self._sample_large(bqm, time_limit=time_limit, **kwargs)

        return self._sample(bqm, time_limit=time_limit, **kwargs)

    def _sample(self, bqm, **kwargs):
        """Sample from the given BQM."""
        with bqm_to_file(bqm, version=2) as fv:
            sapi_problem_id = self.solver.upload_bqm(fv).result()

        return self.solver.sample_bqm(sapi_problem_id, **kwargs).sampleset

    def _sample_large(self, bqm, **kwargs):
        """Sample from the unlabelled version of the BQM, then apply the
        labels to the returned sampleset.
        """
        with bqm_to_file(bqm, version=2, ignore_labels=True) as fv:
            sapi_problem_id = self.solver.upload_bqm(fv).result()

        sampleset = self.solver.sample_bqm(sapi_problem_id, **kwargs).sampleset

        # relabel, as of dimod 0.9.5+ this is not blocking
        mapping = dict(enumerate(bqm.variables))
        return sampleset.relabel_variables(mapping)

    def min_time_limit(self, bqm):
        """Return the minimum `time_limit` accepted for the given problem.

        The minimum time for a hybrid BQM solver is specified as a piecewise-linear
        curve defined by a set of floating-point pairs, the `minimum_time_limit`
        field under :attr:`~dwave.system.samplers.LeapHybridSampler.properties`.
        The first element in each pair is the number of problem variables; the
        second is the minimum required time. The minimum time for any number of
        variables is a linear interpolation calculated on two pairs that represent
        the relevant range for the given number of variables.

        Examples:
            For a solver where
            `LeapHybridSampler().properties["minimum_time_limit"]` returns
            `[[1, 0.1], [100, 10.0], [1000, 20.0]]`, the minimum time for a
            problem 50 variales is 5 seconds (the linear interpolation of the
            first two pairs that represent problems with between 1 to 100
            variables).
        """

        xx, yy = zip(*self.properties["minimum_time_limit"])
        return np.interp([bqm.num_variables], xx, yy)[0]

LeapHybridBQMSampler = LeapHybridSampler


class LeapHybridDQMSampler:
    """A class for using Leap's cloud-based hybrid DQM solvers.

    Leap’s quantum-classical hybrid DQM solvers are intended to solve arbitrary
    application problems formulated as **discrete** quadratic models (DQM).

    You can configure your :term:`solver` selection and usage by setting parameters,
    hierarchically, in a configuration file, as environment variables, or
    explicitly as input arguments, as described in
    `D-Wave Cloud Client <https://docs.ocean.dwavesys.com/en/stable/docs_cloud/sdk_index.html>`_.

    :ref:`dwave-cloud-client <sdk_index_cloud>`'s
    :meth:`~dwave.cloud.client.Client.get_solvers` method filters solvers you have
    access to by `solver properties <https://docs.dwavesys.com/docs/latest/c_solver_properties.html>`_
    ``category=hybrid`` and ``supported_problem_type=dqm``. By default, online
    hybrid DQM solvers are returned ordered by latest ``version``.

    The default specification for filtering and ordering solvers by features is
    available as :attr:`.default_solver` property. Explicitly specifying a
    solver in a configuration file, an environment variable, or keyword
    arguments overrides this specification. See the example in :class:`.LeapHybridSampler`
    on how to extend it instead.

    Args:
        **config:
            Keyword arguments passed to :meth:`dwave.cloud.client.Client.from_config`.

    Examples:
        This example solves a small, illustrative problem: a game of
        rock-paper-scissors. The DQM has two variables representing two hands,
        with cases for rock, paper, scissors. Quadratic biases are set to
        produce a lower value of the DQM for cases of variable ``my_hand``
        interacting with cases of variable ``their_hand`` such that the former
        wins over the latter; for example, the interaction of ``rock-scissors`` is
        set to -1 while ``scissors-rock`` is set to +1.

        >>> import dimod
        >>> from dwave.system import LeapHybridDQMSampler
        ...
        >>> cases = ["rock", "paper", "scissors"]
        >>> win = {"rock": "scissors", "paper": "rock", "scissors": "paper"}
        ...
        >>> dqm = dimod.DiscreteQuadraticModel()
        >>> dqm.add_variable(3, label='my_hand')
        'my_hand'
        >>> dqm.add_variable(3, label='their_hand')
        'their_hand'
        >>> for my_idx, my_case in enumerate(cases):
        ...    for their_idx, their_case in enumerate(cases):
        ...       if win[my_case] == their_case:
        ...          dqm.set_quadratic('my_hand', 'their_hand',
        ...                            {(my_idx, their_idx): -1})
        ...       if win[their_case] == my_case:
        ...          dqm.set_quadratic('my_hand', 'their_hand',
        ...                            {(my_idx, their_idx): 1})
        ...
        >>> dqm_sampler = LeapHybridDQMSampler()      # doctest: +SKIP
        ...
        >>> sampleset = dqm_sampler.sample_dqm(dqm)   # doctest: +SKIP
        >>> print("{} beats {}".format(cases[sampleset.first.sample['my_hand']],
        ...                            cases[sampleset.first.sample['their_hand']]))   # doctest: +SKIP
        rock beats scissors
    """

    @classproperty
    def default_solver(self):
        """dict: Features used to select the latest accessible hybrid DQM solver."""
        return dict(supported_problem_types__contains='dqm',
                    order_by='-properties.version')

    def __init__(self, **config):
        # strongly prefer hybrid solvers; requires kwarg-level override
        config.setdefault('client', 'hybrid')

        # default to short-lived session to prevent resets on slow uploads
        config.setdefault('connection_close', True)

        if FeatureFlags.hss_solver_config_override:
            # use legacy behavior (override solver config from env/file)
            solver = config.setdefault('solver', {})
            if isinstance(solver, abc.Mapping):
                solver.update(self.default_solver)

        # prefer the latest hybrid DQM solver available, but allow for an easy
        # override on any config level above the defaults (file/env/kwarg)
        defaults = config.setdefault('defaults', {})
        if not isinstance(defaults, abc.Mapping):
            raise TypeError("mapping expected for 'defaults'")
        defaults.update(solver=self.default_solver)

        self.client = Client.from_config(**config)
        self.solver = self.client.get_solver()

        # check user-specified solver conforms to our requirements
        if self.properties.get('category') != 'hybrid':
            raise ValueError("selected solver is not a hybrid solver.")
        if 'dqm' not in self.solver.supported_problem_types:
            raise ValueError("selected solver does not support the 'dqm' problem type.")

    @property
    def properties(self) -> Dict[str, Any]:
        """Solver properties as returned by a SAPI query.

        `Solver properties <https://docs.dwavesys.com/docs/latest/c_solver_properties.html>`_
        are dependent on the selected solver and subject to change.
        """
        try:
            return self._properties
        except AttributeError:
            self._properties = properties = self.solver.properties.copy()
            return properties

    @property
    def parameters(self) -> Dict[str, list]:
        """Solver parameters in the form of a dict, where keys
        are keyword parameters accepted by a SAPI query and values are lists of
        properties in
        :attr:`~dwave.system.samplers.LeapHybridDQMSampler.properties` for each
        key.

        `Solver parameters <https://docs.dwavesys.com/docs/latest/c_solver_parameters.html>`_
        are dependent on the selected solver and subject to change.
        """
        try:
            return self._parameters
        except AttributeError:
            parameters = {param: ['parameters']
                          for param in self.properties['parameters']}
            parameters.update(label=[])
            self._parameters = parameters
            return parameters

    @dimod.decorators.nonblocking_sample_method
    def sample_dqm(self, dqm, time_limit=None, compress=False, compressed=None, **kwargs):
        """Sample from the specified discrete quadratic model.

        Args:
            dqm (:obj:`dimod.DiscreteQuadraticModel`):
                Discrete quadratic model (DQM).

                Note that if `dqm` is a :class:`dimod.CaseLabelDQM`, then
                :meth:`~dimod.CaseLabelDQM.map_sample` will need to be used to
                restore the case labels in the returned sample set.

            time_limit (int, optional):
                Maximum run time, in seconds, to allow the solver to work on the
                problem. Must be at least the minimum required for the number of
                problem variables, which is calculated and set by default.

                :meth:`~dwave.system.samplers.LeapHybridDQMSampler.min_time_limit`
                calculates (and describes) the minimum time for your problem.

            compress (binary, optional):
                Compresses the DQM data when set to True. Use if your problem
                somewhat exceeds the maximum allowed size. Compression tends to
                be slow and more effective on homogenous data, which in this
                case means it is more likely to help on DQMs with many identical
                integer-valued biases than ones with random float-valued biases,
                for example.

            compressed (binary, optional):
                Deprecated; please use ``compress`` instead.

            **kwargs:
                Optional keyword arguments for the solver, specified in
                :attr:`~dwave.system.samplers.LeapHybridDQMSampler.parameters`.

        Returns:
            :class:`~dimod.SampleSet`: Sample set constructed from a (non-blocking)
            :class:`~concurrent.futures.Future`-like object.

        Examples:
            See the example in :class:`LeapHybridDQMSampler`.

        """
        if time_limit is None:
            time_limit = self.min_time_limit(dqm)
        elif time_limit < self.min_time_limit(dqm):
            raise ValueError("the minimum time limit is {}s ({}s provided)"
                             "".format(self.min_time_limit(dqm), time_limit))

        # check the max time_limit if it's available
        if 'maximum_time_limit_hrs' in self.properties:
            if time_limit > 60*60*self.properties['maximum_time_limit_hrs']:
                raise ValueError("time_limit cannot exceed the solver maximum "
                                 "of {} hours ({} seconds given)".format(
                                    self.properties['maximum_time_limit_hrs'],
                                    time_limit))

        # we convert to a file here rather than let the cloud-client handle
        # it because we want to strip the labels and let the user handle
        # note: SpooledTemporaryFile currently returned by DQM.to_file
        # does not implement io.BaseIO interface, so we use the underlying
        # (and internal) file-like object for now

        if compressed is not None:
            warn(
                "Argument 'compressed' is deprecated and in future will raise an "
                "exception; please use 'compress' instead.",
                DeprecationWarning, stacklevel=2
                )
            compress = compressed or compress

        try:
            f = dqm.to_file(compress=compress, ignore_labels=True)._file
        except NotImplementedError:
            f = dimod.DQM.to_file(dqm, compress=compress, ignore_labels=True)._file

        future = self.solver.sample_dqm(f, time_limit=time_limit, **kwargs)
        yield future

        sampleset = future.sampleset.relabel_variables(dict(enumerate(dqm.variables)))

        if hasattr(dqm, 'offset') and dqm.offset:
            # dimod 0.10+
            # some versions of HSS don't account for the offset and it's hard
            # to tell which
            sampleset.record.energy = dqm.energies(sampleset)

        yield sampleset

    def min_time_limit(self, dqm):
        """Return the minimum `time_limit` accepted for the given problem.

        The minimum time for a hybrid DQM solver is specified as a
        piecewise-linear curve defined by a set of floating-point pairs,
        the `minimum_time_limit` field under
        :attr:`~dwave.system.samplers.LeapHybridDQMSampler.properties`.
        The first element in each pair is a combination of the numbers of
        interactions, variables, and cases that reflects the "density" of
        connectivity between the problem's variables;
        the second is the minimum required time. The minimum time for any
        particular problem size is a linear interpolation calculated on
        two pairs that represent the relevant range for the given problem.

        Examples:
            For a solver where
            `LeapHybridDQMSampler().properties["minimum_time_limit"]` returns
            `[[1, 0.1], [100, 10.0], [1000, 20.0]]`, the minimum time for a
            problem of "density" 50 is 5 seconds (the linear interpolation of the
            first two pairs that represent problems with "density" between 1 to
            100).
        """
        ec = (dqm.num_variable_interactions() * dqm.num_cases() /
              max(dqm.num_variables(), 1))
        limits = np.array(self.properties['minimum_time_limit'])
        t = np.interp(ec, limits[:, 0], limits[:, 1])
        return max([5, t])


class LeapHybridCQMSampler:
    """A class for using Leap's cloud-based hybrid CQM solvers.

    Leap’s quantum-classical hybrid CQM solvers are intended to solve
    application problems formulated as
    :ref:`constrained quadratic models (CQM) <cqm_sdk>`.

    You can configure your :term:`solver` selection and usage by setting parameters,
    hierarchically, in a configuration file, as environment variables, or
    explicitly as input arguments, as described in
    `D-Wave Cloud Client <https://docs.ocean.dwavesys.com/en/stable/docs_cloud/sdk_index.html>`_.

    :ref:`dwave-cloud-client <sdk_index_cloud>`'s
    :meth:`~dwave.cloud.client.Client.get_solvers` method filters solvers you have
    access to by `solver properties <https://docs.dwavesys.com/docs/latest/c_solver_properties.html>`_
    ``category=hybrid`` and ``supported_problem_type=cqm``. By default, online
    hybrid CQM solvers are returned ordered by latest ``version``.

    Args:
        **config:
            Keyword arguments passed to :meth:`dwave.cloud.client.Client.from_config`.

    Examples:
        This example solves a simple problem of finding the rectangle with the
        greatest area when the perimeter is limited. In this example, the
        perimeter of the rectangle is set to 8 (meaning the largest area is for
        the :math:`2X2` square).

        A CQM is created that will have two integer variables, :math:`i, j`, each
        limited to half the maximum perimeter length of 8, to represent the
        lengths of the rectangle's sides:

        >>> from dimod import ConstrainedQuadraticModel, Integer
        >>> i = Integer('i', upper_bound=4)
        >>> j = Integer('j', upper_bound=4)
        >>> cqm = ConstrainedQuadraticModel()

        The area of the rectangle is given by the multiplication of side :math:`i`
        by side :math:`j`. The goal is to maximize the area, :math:`i*j`. Because
        D-Wave samplers minimize, the objective should have its lowest value when
        this goal is met. Objective :math:`-i*j` has its minimum value when
        :math:`i*j`, the area, is greatest:

        >>> cqm.set_objective(-i*j)

        Finally, the requirement that the sum of both sides must not exceed the
        perimeter is represented as constraint :math:`2i + 2j <= 8`:

        >>> cqm.add_constraint(2*i+2*j <= 8, "Max perimeter")
        'Max perimeter'

        Instantiate a hybrid CQM sampler and submit the problem for solution by
        a remote solver provided by the Leap quantum cloud service:

        >>> from dwave.system import LeapHybridCQMSampler   # doctest: +SKIP
        >>> sampler = LeapHybridCQMSampler()                # doctest: +SKIP
        >>> sampleset = sampler.sample_cqm(cqm)             # doctest: +SKIP
        >>> print(sampleset.first)                          # doctest: +SKIP
        Sample(sample={'i': 2.0, 'j': 2.0}, energy=-4.0, num_occurrences=1,
        ...            is_feasible=True, is_satisfied=array([ True]))

        The best (lowest-energy) solution found has :math:`i=j=2` as expected,
        a solution that is feasible because all the constraints (one in this
        example) are satisfied.

    """
    def __init__(self, **config):
        # strongly prefer hybrid solvers; requires kwarg-level override
        config.setdefault('client', 'hybrid')

        # default to short-lived session to prevent resets on slow uploads
        config.setdefault('connection_close', True)

        if FeatureFlags.hss_solver_config_override:
            # use legacy behavior (override solver config from env/file)
            solver = config.setdefault('solver', {})
            if isinstance(solver, abc.Mapping):
                solver.update(self.default_solver)

        # prefer the latest hybrid CQM solver available, but allow for an easy
        # override on any config level above the defaults (file/env/kwarg)
        defaults = config.setdefault('defaults', {})
        if not isinstance(defaults, abc.Mapping):
            raise TypeError("mapping expected for 'defaults'")
        defaults.update(solver=self.default_solver)

        self.client = Client.from_config(**config)
        self.solver = self.client.get_solver()

        # For explicitly named solvers:
        if self.properties.get('category') != 'hybrid':
            raise ValueError("selected solver is not a hybrid solver.")
        if 'cqm' not in self.solver.supported_problem_types:
            raise ValueError("selected solver does not support the 'cqm' problem type.")

    @classproperty
    def default_solver(cls) -> Dict[str, str]:
        """Features used to select the latest accessible hybrid CQM solver."""
        return dict(supported_problem_types__contains='cqm',
                    order_by='-properties.version')

    @property
    def properties(self) -> Dict[str, Any]:
        """Solver properties as returned by a SAPI query.

        `Solver properties <https://docs.dwavesys.com/docs/latest/c_solver_properties.html>`_
        are dependent on the selected solver and subject to change.
        """
        try:
            return self._properties
        except AttributeError:
            self._properties = properties = self.solver.properties.copy()
            return properties

    @property
    def parameters(self) -> Dict[str, List[str]]:
        """Solver parameters in the form of a dict, where keys
        are keyword parameters accepted by a SAPI query and values are lists of
        properties in
        :attr:`~dwave.system.samplers.LeapHybridCQMSampler.properties` for each
        key.

        `Solver parameters <https://docs.dwavesys.com/docs/latest/c_solver_parameters.html>`_
        are dependent on the selected solver and subject to change.
        """
        try:
            return self._parameters
        except AttributeError:
            parameters = {param: ['parameters']
                          for param in self.properties['parameters']}
            parameters.update(label=[])
            self._parameters = parameters
            return parameters

    def sample_cqm(self, cqm: dimod.ConstrainedQuadraticModel,
                   time_limit: Optional[float] = None, **kwargs):
        """Sample from the specified constrained quadratic model.

        Args:
            cqm (:obj:`dimod.ConstrainedQuadraticModel`):
                Constrained quadratic model (CQM).

            time_limit (int, optional):
                Maximum run time, in seconds, to allow the solver to work on the
                problem. Must be at least the minimum required for the problem,
                which is calculated and set by default.

                :meth:`~dwave.system.samplers.LeapHybridCQMSampler.min_time_limit`
                calculates (and describes) the minimum time for your problem.

            **kwargs:
                Optional keyword arguments for the solver, specified in
                :attr:`~dwave.system.samplers.LeapHybridCQMSampler.parameters`.

        Returns:
            :class:`~dimod.SampleSet`: Sample set constructed from a (non-blocking)
            :class:`~concurrent.futures.Future`-like object.

        Examples:
            See the example in :class:`LeapHybridCQMSampler`.

        """

        if not isinstance(cqm, dimod.ConstrainedQuadraticModel):
            raise TypeError("first argument 'cqm' must be a ConstrainedQuadraticModel, "
                            f"recieved {type(cqm).__name__}")

        if time_limit is None:
            time_limit = self.min_time_limit(cqm)
        elif time_limit < self.min_time_limit(cqm):
            raise ValueError("the minimum time limit for this problem is "
                             f"{self.min_time_limit(cqm)} seconds "
                             f"({time_limit}s provided), "
                             "see .min_time_limit method")

        contact_sales_str = "Contact D-Wave at sales@dwavesys.com if your " + \
                            "application requires scale or performance that " + \
                            "exceeds the currently advertised capabilities of " + \
                            "this hybrid solver."

        if len(cqm.constraints) > self.properties['maximum_number_of_constraints']:
            raise ValueError(
                "constrained quadratic model must have "
                f"{self.properties['maximum_number_of_constraints']} or fewer "
                f"constraints; given model has {len(cqm.constraints)}. "
                f"{contact_sales_str}")

        if len(cqm.variables) > self.properties['maximum_number_of_variables']:
            raise ValueError(
                "constrained quadratic model must have "
                f"{self.properties['maximum_number_of_variables']} or fewer "
                f"variables; given model has {len(cqm.variables)}. "
                f"{contact_sales_str}")

        if cqm.num_biases() > self.properties['maximum_number_of_biases']:
            raise ValueError(
                "constrained quadratic model must have "
                f"{self.properties['maximum_number_of_biases']} or fewer "
                f"biases; given model has {cqm.num_biases()}. "
                f"{contact_sales_str}")

        if cqm.num_quadratic_variables() > self.properties['maximum_number_of_quadratic_variables']:
            raise ValueError(
                "constrained quadratic model must have "
                f"{self.properties['maximum_number_of_quadratic_variables']} "
                "or fewer variables with at least one quadratic bias across "
                "all constraints; given model has "
                f"{cqm.num_quadratic_variables()}. "
                f"{contact_sales_str}")

        return self.solver.sample_cqm(cqm, time_limit=time_limit, **kwargs).sampleset

    def min_time_limit(self, cqm: dimod.ConstrainedQuadraticModel) -> float:
        """Return the minimum `time_limit` accepted for the given problem."""

        # todo: remove the hard-coded defaults
        num_variables_multiplier = self.properties.get('num_variables_multiplier', 1.57e-04)
        num_biases_multiplier = self.properties.get('num_biases_multiplier', 4.65e-06)
        num_constraints_multiplier = self.properties.get('num_constraints_multiplier', 6.44e-09)
        minimum_time_limit = self.properties['minimum_time_limit_s']

        num_variables = len(cqm.variables)
        num_constraints = len(cqm.constraints)
        num_biases = cqm.num_biases()

        return max(
            num_variables_multiplier * num_variables +
            num_biases_multiplier * num_biases +
            num_constraints_multiplier * num_variables * num_constraints,
            minimum_time_limit
            )



# Copyright 2018 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import unittest
import itertools
import random

from collections.abc import Mapping

import networkx as nx
import numpy as np
import numpy.testing as npt

from parameterized import parameterized

import dimod

import dwave.embedding


class TestUnembedSampleSet(unittest.TestCase):

    def test_majority_vote(self):
        """should return the most common value in the chain"""

        sample0 = {0: -1, 1: -1, 2: +1}
        sample1 = {0: +1, 1: -1, 2: +1}
        samples = [sample0, sample1]

        embedding = {'a': {0, 1, 2}}

        bqm = dimod.BinaryQuadraticModel.from_ising({'a': 1}, {})

        resp = dimod.SampleSet.from_samples(samples, energy=[-1, 1], info={}, vartype=dimod.SPIN)

        resp = dwave.embedding.unembed_sampleset(resp, embedding, bqm, chain_break_method=dwave.embedding.majority_vote)

        # specify that majority vote should be used
        source_samples = list(resp)

        self.assertEqual(source_samples, [{'a': -1}, {'a': +1}])

    def test_majority_vote_with_response(self):
        sample0 = {0: -1, 1: -1, 2: +1}
        sample1 = {0: +1, 1: -1, 2: +1}
        samples = [sample0, sample1]

        # now set up an embedding that works for one sample and not the other
        embedding = {'a': {0, 1, 2}}

        bqm = dimod.BinaryQuadraticModel.from_ising({'a': 1}, {})

        resp = dimod.SampleSet.from_samples(samples, energy=[-1, 1], info={}, vartype=dimod.SPIN)

        resp = dwave.embedding.unembed_sampleset(resp, embedding, bqm, chain_break_method=dwave.embedding.majority_vote)

        # specify that majority vote should be used
        source_samples = list(resp)

        self.assertEqual(len(source_samples), 2)
        self.assertEqual(source_samples, [{'a': -1}, {'a': +1}])

    def test_discard(self):
        sample0 = {0: -1, 1: -1, 2: +1}
        sample1 = {0: +1, 1: -1, 2: +1}
        samples = [sample0, sample1]

        embedding = {'a': {0, 1, 2}}

        bqm = dimod.BinaryQuadraticModel.from_ising({'a': 1}, {})

        resp = dimod.SampleSet.from_samples(samples, energy=[-1, 1], info={}, vartype=dimod.SPIN)

        resp = dwave.embedding.unembed_sampleset(resp, embedding, bqm, chain_break_method=dwave.embedding.discard)

        source_samples = list(resp)

        # no samples should be returned because they are all broken
        self.assertEqual(len(source_samples), 0)

        # now set up an embedding that works for one sample and not the other
        embedding = {'a': {0, 1}, 'b': {2}}

        bqm = dimod.BinaryQuadraticModel.from_ising({'a': 1}, {('a', 'b'): -1})

        resp = dimod.SampleSet.from_samples(samples, energy=[-1, 1], info={}, vartype=dimod.SPIN)

        resp = dwave.embedding.unembed_sampleset(resp, embedding, bqm, chain_break_method=dwave.embedding.discard)

        source_samples = list(resp)
        # only the first sample should be returned
        self.assertEqual(len(source_samples), 1)
        self.assertEqual(source_samples, [{'a': -1, 'b': +1}])

    def test_discard_with_response(self):
        sample0 = {0: -1, 1: -1, 2: +1}
        sample1 = {0: +1, 1: -1, 2: +1}
        samples = [sample0, sample1]

        # now set up an embedding that works for one sample and not the other
        embedding = {'a': {0, 1}, 'b': {2}}

        bqm = dimod.BinaryQuadraticModel.from_ising({'a': 1}, {('a', 'b'): -1})

        resp = dimod.SampleSet.from_samples(samples, energy=[-1, 1], info={}, vartype=dimod.SPIN)

        resp = dwave.embedding.unembed_sampleset(resp, embedding, bqm, chain_break_method=dwave.embedding.discard)

        source_samples = list(resp)

        # only the first sample should be returned
        self.assertEqual(len(source_samples), 1)
        self.assertEqual(source_samples, [{'a': -1, 'b': +1}])

    def test_energies_functional(self):
        h = {'a': .1, 'b': 0, 'c': 0}
        J = {('a', 'b'): 1, ('b', 'c'): 1.3, ('a', 'c'): -1}
        bqm = dimod.BinaryQuadraticModel.from_ising(h, J, offset=1.3)

        embedding = {'a': {0}, 'b': {1}, 'c': {2, 3}}

        embedded_bqm = dwave.embedding.embed_bqm(bqm, embedding, nx.cycle_graph(4), chain_strength=1)

        embedded_response = dimod.ExactSolver().sample(embedded_bqm)

        response = dwave.embedding.unembed_sampleset(embedded_response, embedding, bqm)

        for sample, energy in response.data(['sample', 'energy']):
            self.assertAlmostEqual(bqm.energy(sample), energy)

    def test_energies_discard(self):
        h = {'a': .1, 'b': 0, 'c': 0}
        J = {('a', 'b'): 1, ('b', 'c'): 1.3, ('a', 'c'): -1}
        bqm = dimod.BinaryQuadraticModel.from_ising(h, J, offset=1.3)

        embedding = {'a': {0}, 'b': {1}, 'c': {2, 3}}

        embedded_bqm = dwave.embedding.embed_bqm(bqm, embedding, nx.cycle_graph(4), chain_strength=1)

        embedded_response = dimod.ExactSolver().sample(embedded_bqm)

        chain_break_method = dwave.embedding.discard
        response = dwave.embedding.unembed_sampleset(embedded_response, embedding, bqm,
                                                   chain_break_method=chain_break_method)

        self.assertEqual(len(embedded_response) / 2, len(response))  # half chains should be broken

        for sample, energy in response.data(['sample', 'energy']):
            self.assertEqual(bqm.energy(sample), energy)

    def test_unembed_sampleset_with_discard_matrix_typical(self):
        h = {'a': .1, 'b': 0, 'c': 0}
        J = {('a', 'b'): 1, ('b', 'c'): 1.3, ('a', 'c'): -1}
        bqm = dimod.BinaryQuadraticModel.from_ising(h, J, offset=1.3)

        embedding = {'a': {0}, 'b': {1}, 'c': {2, 3}}

        embedded_bqm = dwave.embedding.embed_bqm(bqm, embedding, nx.cycle_graph(4), chain_strength=1)

        embedded_response = dimod.ExactSolver().sample(embedded_bqm)

        chain_break_method = dwave.embedding.discard
        response = dwave.embedding.unembed_sampleset(embedded_response, embedding, bqm,
                                                   chain_break_method=dwave.embedding.discard)

        self.assertEqual(len(embedded_response) / 2, len(response))  # half chains should be broken

        for sample, energy in response.data(['sample', 'energy']):
            self.assertEqual(bqm.energy(sample), energy)

    def test_embedding_superset(self):
        # source graph in the embedding is a superset of the bqm
        response = dimod.SampleSet(np.rec.array([([-1,  1, -1,  1, -1,  1, -1,  1], -1.4, 1),
                                                ([-1,  1, -1, -1, -1,  1, -1, -1], -1.4, 1),
                                                ([+1, -1, -1, -1,  1, -1, -1, -1], -1.6, 1),
                                                ([+1, -1, -1, -1,  1, -1, -1, -1], -1.6, 1)],
                                   dtype=[('sample', 'i1', (8,)), ('energy', '<f8'), ('num_occurrences', '<i8')]),
                                   [0, 1, 2, 3, 4, 5, 6, 7], {}, 'SPIN')
        embedding = {0: {0, 4}, 1: {1, 5}, 2: {2, 6}, 3: {3, 7}}
        bqm = dimod.BinaryQuadraticModel.from_ising([.1, .2], {(0, 1): 1.5}, 0.0)

        unembedded = dwave.embedding.unembed_sampleset(response, embedding, source_bqm=bqm)

        arr = np.rec.array([([-1,  1], -1.4, 1), ([-1,  1], -1.4, 1), ([+1, -1], -1.6, 1), ([+1, -1], -1.6, 1)],
                           dtype=[('sample', 'i1', (2,)), ('energy', '<f8'), ('num_occurrences', '<i8')])

        np.testing.assert_array_equal(arr, unembedded.record)

    def test_chain_break_statistics_discard(self):
        sample0 = {0: -1, 1: -1, 2: +1}
        sample1 = {0: +1, 1: -1, 2: +1}
        samples = [sample0, sample1]

        # now set up an embedding that works for one sample and not the other
        embedding = {'a': {0, 1}, 'b': {2}}

        bqm = dimod.BinaryQuadraticModel.from_ising({'a': 1}, {('a', 'b'): -1})

        resp = dimod.SampleSet.from_samples(samples, energy=[-1, 1], info={}, vartype=dimod.SPIN)

        resp = dwave.embedding.unembed_sampleset(resp, embedding, bqm, chain_break_method=dwave.embedding.discard,
                                               chain_break_fraction=True)

        source_samples = list(resp)

        # only the first sample should be returned
        self.assertEqual(len(source_samples), 1)
        self.assertEqual(source_samples, [{'a': -1, 'b': +1}])

        self.assertEqual(resp.record.chain_break_fraction.sum(), 0)

    def test_multi(self):

        samples = [{'a': -1, 'b': -1, 'c': +1, 'd': -1},
                   {'a': -1, 'b': -1, 'c': -1, 'd': +1}]
        embedding = {0: ['a', 'b', 'c'], 1: ['d']}
        bqm = dimod.BinaryQuadraticModel.from_ising({}, {(0, 1): 1})

        resp = dimod.SampleSet.from_samples(samples, energy=[-1, 1], info={},
                                            vartype=dimod.SPIN)

        methods = [dwave.embedding.majority_vote,
                   dwave.embedding.discard]

        ss = dwave.embedding.unembed_sampleset(resp, embedding, bqm,
                                               chain_break_method=methods)

        np.testing.assert_array_equal(ss.record.chain_break_method, [0, 0, 1])


class TestEmbedBQM(unittest.TestCase):
    def test_embed_bqm_empty(self):
        bqm = dimod.BinaryQuadraticModel.empty(dimod.SPIN)

        embedded_bqm = dwave.embedding.embed_bqm(bqm, {}, {})

        self.assertIsInstance(embedded_bqm, dimod.BinaryQuadraticModel)
        self.assertFalse(embedded_bqm)  # should be empty

    @parameterized.expand(
        [('_'.join(vt.name for vt in vartypes), *vartypes) for vartypes
         in itertools.product([dimod.BINARY, dimod.SPIN], repeat=3)])
    def test_change_vartype_return(self, name,
                                   source_vartype, smear_vartype,
                                   target_vartype):

        bqm = dimod.BinaryQuadraticModel.empty(source_vartype)
        embedded_bqm = dwave.embedding.embed_bqm(bqm, {}, {},
                                                 smear_vartype=smear_vartype)

        self.assertIs(embedded_bqm.vartype, source_vartype)
        embedded_bqm.change_vartype(target_vartype, inplace=False)
        embedded_bqm.change_vartype(target_vartype, inplace=True)

    def test_embed_bqm_subclass_propagation(self):

        class MyBQM(dimod.BinaryQuadraticModel):
            pass

        bqm = MyBQM.empty(dimod.BINARY)

        embedded_bqm = dwave.embedding.embed_bqm(bqm, {}, {})

        self.assertIsInstance(embedded_bqm, dimod.BinaryQuadraticModel)
        self.assertIsInstance(embedded_bqm, MyBQM)
        self.assertFalse(embedded_bqm)  # should be empty

    def test_embed_bqm_only_offset(self):
        bqm = dimod.BinaryQuadraticModel({}, {}, 1.0, dimod.SPIN)

        embedded_bqm = dwave.embedding.embed_bqm(bqm, {}, {})

        self.assertEqual(bqm, embedded_bqm)

    def test_embed_bqm_identity(self):

        bqm = dimod.BinaryQuadraticModel({'a': -1}, {(0, 1): .5, (1, 'a'): -1.}, 1.0, dimod.BINARY)

        embedding = {v: {v} for v in bqm.linear}  # identity embedding

        target_adj = nx.Graph()
        target_adj.add_nodes_from(bqm.linear)
        target_adj.add_edges_from(bqm.quadratic)

        embedded_bqm = dwave.embedding.embed_bqm(bqm, embedding, target_adj)

        self.assertEqual(bqm, embedded_bqm)

    def test_embed_bqm_NAE3SAT_to_square(self):

        h = {'a': 0, 'b': 0, 'c': 0}
        J = {('a', 'b'): 1, ('b', 'c'): 1, ('a', 'c'): 1}

        bqm = dimod.BinaryQuadraticModel.from_ising(h, J)

        embedding = {'a': {0}, 'b': {1}, 'c': {2, 3}}

        embedded_bqm = dwave.embedding.embed_bqm(bqm, embedding, nx.cycle_graph(4), chain_strength=1)

        self.assertEqual(embedded_bqm,
                         dimod.BinaryQuadraticModel({0: 0, 1: 0, 2: 0, 3: 0},
                                                    {(0, 1): 1, (1, 2): 1, (2, 3): -1, (0, 3): 1},
                                                    1.0,  # offset the energy from satisfying chains
                                                    dimod.SPIN))

        # check that the energy has been preserved
        for config in itertools.product((-1, 1), repeat=3):
            sample = dict(zip(('a', 'b', 'c'), config))
            target_sample = {u: sample[v] for v, chain in embedding.items() for u in chain}  # no chains broken
            self.assertAlmostEqual(bqm.energy(sample), embedded_bqm.energy(target_sample))

    def test_embed_ising_components_empty(self):
        h = {}
        J = {}
        embedding = {}
        adjacency = {}

        he, Je = dwave.embedding.embed_ising(h, J, embedding, adjacency)

        self.assertFalse(he)
        self.assertFalse(Je)

    def test_embed_ising_bad_chain(self):
        h = {}
        j = {(0, 1): 1}
        embeddings = {0: {0, 2}, 1: {1}}  # (0, 2) not connected
        adj = {0: {1}, 1: {0, 2}, 2: {1}}

        with self.assertRaises(dwave.embedding.exceptions.DisconnectedChainError):
            dwave.embedding.embed_ising(h, j, embeddings, adj)

    def test_embed_ising_nonadj(self):
        """chain in embedding has non-adjacenct nodes"""
        h = {}
        j = {(0, 1): 1}
        embeddings = {0: {0, 1}, 1: {3, 4}}

        adj = nx.Graph()
        adj.add_edges_from({(0, 1), (1, 2), (2, 3), (3, 4)})

        with self.assertRaises(dwave.embedding.exceptions.EmbeddingError):
            dwave.embedding.embed_ising(h, j, embeddings, adj)

    def test_embed_ising_h_embedding_mismatch(self):
        """there is not a mapping in embedding for every var in h"""
        h = dict(zip(range(3), [1, 2, 3]))
        j = {}
        embedding = {0: (0, 1), 1: (2,)}
        adj = nx.Graph()
        adj.add_edges_from({(0, 1), (1, 2)})

        with self.assertRaises(dwave.embedding.exceptions.EmbeddingError):
            dwave.embedding.embed_ising(h, j, embedding, adj)

    def test_embed_ising_j_index_too_large(self):
        """j references a variable not mentioned in embedding"""
        h = {}
        j = {(0, 1): -1, (0, 2): 1}
        embedding = {0: (0, 1), 1: (2,)}
        adj = nx.Graph()
        adj.add_edges_from({(0, 1), (1, 2)})

        with self.assertRaises(dwave.embedding.exceptions.EmbeddingError):
            dwave.embedding.embed_ising(h, j, embedding, adj)

    def test_embed_ising_typical(self):
        h = {0: 1, 1: 10}
        j = {(0, 1): 15, (2, 1): -8, (0, 2): 5, (2, 0): -2}
        embeddings = {0: [1], 1: [2, 3], 2: [0]}

        adj = nx.Graph()
        adj.add_edges_from({(0, 1), (1, 2), (2, 3), (3, 0), (2, 0)})

        expected_h0 = {0: 0, 1: 1, 2: 5, 3: 5}
        expected_j0 = {(0, 1): 3, (0, 2): -4, (0, 3): -4, (1, 2): 15, (2, 3): -19.9302}

        h0, j0 = dwave.embedding.embed_ising(h, j, embeddings, adj)
        self.assertEqual(h0, expected_h0)

        # check j0
        for (u, v), bias in j0.items():
            self.assertTrue((u, v) in expected_j0 or (v, u) in expected_j0)
            self.assertFalse((u, v) in expected_j0 and (v, u) in expected_j0)

            if (u, v) in expected_j0:
                self.assertAlmostEqual(expected_j0[(u, v)], bias, 4)
            else:
                self.assertAlmostEqual(expected_j0[(v, u)], bias, 4)

    def test_embed_ising_embedding_not_in_adj(self):
        """embedding refers to a variable not in the adjacency"""
        h = {0: 0, 1: 0}
        J = {(0, 1): 1}

        embedding = {0: [0, 1], 1: [2]}

        adjacency = {0: {1}, 1: {0}}

        with self.assertRaises(dwave.embedding.exceptions.EmbeddingError):
            dwave.embedding.embed_ising(h, J, embedding, adjacency)

    def test_embed_bqm_BINARY(self):
        Q = {('a', 'a'): 0, ('a', 'b'): -1, ('b', 'b'): 0, ('c', 'c'): 0, ('b', 'c'): -1}
        bqm = dimod.BinaryQuadraticModel.from_qubo(Q)

        embedding = {'a': {0}, 'b': {1}, 'c': {2, 3}}

        embedded_bqm = dwave.embedding.embed_bqm(bqm, embedding, nx.cycle_graph(4), chain_strength=1)

        # check that the energy has been preserved
        for config in itertools.product((0, 1), repeat=3):
            sample = dict(zip(('a', 'b', 'c'), config))
            target_sample = {u: sample[v] for v, chain in embedding.items() for u in chain}  # no chains broken
            self.assertAlmostEqual(bqm.energy(sample), embedded_bqm.energy(target_sample))

    def test_embed_qubo(self):
        Q = {('a', 'a'): 0, ('a', 'b'): -1, ('b', 'b'): 0, ('c', 'c'): 0, ('b', 'c'): -1}
        bqm = dimod.BinaryQuadraticModel.from_qubo(Q)

        embedding = {'a': {0}, 'b': {1}, 'c': {2, 3}}

        target_Q = dwave.embedding.embed_qubo(Q, embedding, nx.cycle_graph(4), chain_strength=1)

        test_Q = {(0, 0): 0.0, (1, 1): 0.0, (2, 2): 2.0, (3, 3): 2.0,
                  (0, 1): -1.0, (1, 2): -1.0, (2, 3): -4.0}

        self.assertEqual(len(target_Q), len(test_Q))
        for (u, v), bias in target_Q.items():
            if (u, v) in test_Q:
                self.assertEqual(bias, test_Q[(u, v)])
            else:
                self.assertEqual(bias, test_Q[(v, u)])

    def test_embedding_with_extra_chains(self):
        embedding = {0: [0, 1], 1: [2], 2: [3]}
        G = nx.cycle_graph(4)

        bqm = dimod.BinaryQuadraticModel.from_qubo({(0, 0): 1})

        target_bqm = dwave.embedding.embed_bqm(bqm, embedding, G)

        for v, c in embedding.items():
            if v in bqm.variables:
                for q in c:
                    self.assertIn(q, target_bqm.variables)
            else:
                for q in c:
                    self.assertNotIn(q, target_bqm.variables)
            

    def test_empty_chain_exception(self):
        embedding = {0: [], 1: [2], 2: [3]}
        G = nx.cycle_graph(4)

        bqm = dimod.BinaryQuadraticModel.from_qubo({(0, 0): 1})

        with self.assertRaises(dwave.embedding.exceptions.MissingChainError):
            dwave.embedding.embed_bqm(bqm, embedding, G)

    def test_energy_range_embedding(self):
        # start with an Ising
        bqm = dimod.BinaryQuadraticModel.from_ising({}, {(0, 1): 1, (1, 2): 1, (0, 2): 1})

        # convert to BINARY
        bqm.change_vartype(dimod.BINARY, inplace=True)

        # embedding
        embedding = {0: ['a', 'b'], 1: ['c'], 2: ['d']}
        graph = nx.cycle_graph(['a', 'b', 'c', 'd'])
        graph.add_edge('a', 'c')

        # embed
        embedded = dwave.embedding.embed_bqm(bqm, embedding, graph, chain_strength=1, smear_vartype=dimod.SPIN)

        preferred = dimod.BinaryQuadraticModel({'a': 0.0, 'b': 0.0, 'c': 0.0, 'd': 0.0},
                                               {('a', 'c'): 0.5, ('b', 'c'): 0.5, ('c', 'd'): 1.0,
                                                ('a', 'd'): 1.0, ('a', 'b'): -1.0}, 1.0, dimod.SPIN)

        self.assertEqual(embedded.spin, preferred)

    def test_deprecation(self):
        with self.assertWarns(DeprecationWarning):
            dwave.embedding.embed_bqm(dimod.BQM.empty('SPIN'),
                                      dwave.embedding.EmbeddedStructure([], {}),
                                      {})


class TestEmbeddedStructure(unittest.TestCase):
    def test_empty_embedding(self):
        a = dwave.embedding.EmbeddedStructure([], {})
        self.assertEqual(a, {})

        b = dwave.embedding.EmbeddedStructure([(0, 1)], {})
        self.assertEqual(b, {})

    def check_edges(self, embedded_structure, inter_edges, chain_edges):
        for u, v in itertools.product(embedded_structure, embedded_structure):
            if u == v:
                check = chain_edges[u]
                got = list(embedded_structure.chain_edges(u))
                self.assertEqual(check, got)
            else:
                check = inter_edges[u, v]
                got = list(embedded_structure.interaction_edges(u, v))
                self.assertEqual(check, got)

    def test_triangle_to_triangle(self):
        g = [(0, 1), (1, 2), (2, 0)]
        emb = {0: (1,), 1: (2,), 2: (0,)}
        a = dwave.embedding.EmbeddedStructure(g, emb)
        inter_edges = {(0, 1): [(1, 2)], (1, 0): [(2, 1)],
                       (0, 2): [(1, 0)], (2, 0): [(0, 1)],
                       (1, 2): [(2, 0)], (2, 1): [(0, 2)]}
        chain_edges = {i: [] for i in range(3)}
        self.check_edges(a, inter_edges, chain_edges)

        self.check_edges(a.copy(), inter_edges, chain_edges)


    def test_hexagon_to_triangle(self):
        g = [(0, 1), (1, 2), (2, 3), (3, 4), (4, 5), (5, 0)]
        emb = {0: (1, 2), 1: (3, 4), 2: (5, 0)}
        a = dwave.embedding.EmbeddedStructure(g, emb)
        inter_edges = {(0, 1): [(2, 3)], (1, 0): [(3, 2)],
                       (0, 2): [(1, 0)], (2, 0): [(0, 1)],
                       (1, 2): [(4, 5)], (2, 1): [(5, 4)]}

        chain_edges = {u: [c] for u, c in emb.items()}
        self.check_edges(a, inter_edges, chain_edges)

        self.check_edges(a.copy(), inter_edges, chain_edges)

    def test_immutable(self):
        a = dwave.embedding.EmbeddedStructure([], {})
        with self.assertRaises(TypeError):
            a[0] = 0

        with self.assertRaises(TypeError):
            del a[0]

        with self.assertRaises(TypeError):
            a.clear()

        with self.assertRaises(TypeError):
            a.pop()

        with self.assertRaises(TypeError):
            a.popitem()

        with self.assertRaises(TypeError):
            a.setdefault(5, None)

        with self.assertRaises(TypeError):
            a.update((2, 3))

        with self.assertRaises(NotImplementedError):
            a.fromkeys((1,))
    
    def test_disconnected_chain(self):
        g = [(0, 1), (1, 2), (2, 3), (3, 4), (4, 5)]
        emb = {0: (1, 2), 1: (3, 4), 2: (5, 0)}
        with self.assertRaises(dwave.embedding.exceptions.DisconnectedChainError):
            a = dwave.embedding.EmbeddedStructure(g, emb)

    def test_missing_interaction(self):
        g = [(0, 1), (1, 2)]
        emb = {0: (1,), 1: (2,), 2: (0,)}
        a = dwave.embedding.EmbeddedStructure(g, emb)
        inter_edges = {(0, 1): [(1, 2)], (1, 0): [(2, 1)],
                       (0, 2): [(1, 0)], (2, 0): [(0, 1)],
                       (1, 2): [], (2, 1): []}
        chain_edges = {i: [] for i in range(3)}
        self.check_edges(a, inter_edges, chain_edges)

    def test_embed_bqm(self):
        #octahedron
        g = [(0, 1), (1, 2), (2, 3), (3, 4), (4, 5), (5, 0),
             (0, 2), (1, 3), (2, 4), (3, 5), (4, 0), (5, 1)]
        emb = {'a': (1, 2), 'b': (3, 4), 'c': (5, 0)}
        emb_s = dwave.embedding.EmbeddedStructure(g, emb)
        chain_strength = {'a': 10, 'b': 20, 'c': 30}
        linear = {'a': -1, 'b': -2, 'c': -3}
        quadratic = {('a', 'b'): 1, ('a', 'c'): -1, ('b', 'c'): 2}
        source_bqm = dimod.BQM(linear, quadratic, 5, dimod.SPIN)

        goal_linear = {0: -3/2, 1: -1/2, 2: -1/2, 3: -1, 4: -1, 5: -3/2}
        goal_quadratic = {
            (0, 5): -30, (1, 2): -10, (3, 4): -20, #chain edges
            (1, 3): 1/3, (2, 3): 1/3, (2, 4): 1/3, #a-b
            (0, 1): -1/3, (0, 2): -1/3, (1, 5): -1/3, #a-c
            (3, 5): 2/3, (4, 5): 2/3, (4, 0): 2/3, #b-c
        }
        goal_offset = 5 + 10 + 20 + 30
        goal_bqm = dimod.BQM(goal_linear, goal_quadratic, goal_offset, dimod.SPIN)

        target_bqm = emb_s.embed_bqm(source_bqm, chain_strength=chain_strength)

        dimod.testing.assert_bqm_almost_equal(target_bqm, goal_bqm)

        #test the chain_strength is a function case:
        def strength_func(S, E):
            return chain_strength

        target_bqm = emb_s.embed_bqm(source_bqm, chain_strength=strength_func)

        dimod.testing.assert_bqm_almost_equal(target_bqm, goal_bqm)

    def test_copy(self):
        G = [(0, 1), (1, 2)]
        embedding = {0: (1,), 1: (2,), 2: (0,)}
        embedded_structure = dwave.embedding.EmbeddedStructure(G, embedding)

        import copy

        self.assertEqual(embedded_structure.copy(), embedded_structure)
        self.assertEqual(copy.copy(embedded_structure), embedded_structure)
        self.assertEqual(copy.deepcopy(embedded_structure), embedded_structure)

        # deep with memo
        memo = {}
        self.assertIs(copy.deepcopy(embedded_structure, memo),
                      copy.deepcopy(embedded_structure, memo))



# Copyright 2018 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import unittest
import os
from unittest import mock

import numpy

import dimod
from dwave.cloud.exceptions import ConfigFileError, SolverNotFoundError
from dwave.cloud.client import Client
from dwave_networkx.generators.pegasus import pegasus_graph

from dwave.system.samplers import DWaveSampler


@unittest.skipIf(os.getenv('SKIP_INT_TESTS'), "Skipping integration test.")
class TestDWaveSampler(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        try:
            cls.qpu = DWaveSampler()
        except (ValueError, ConfigFileError, SolverNotFoundError):
            raise unittest.SkipTest("no qpu available")

    @classmethod
    def tearDownClass(cls):
        cls.qpu.client.close()

    @staticmethod
    def nonzero_ising_problem(sampler):
        max_h = max(sampler.properties.get('h_range', [-1, 1]))
        max_j = max(sampler.properties.get('j_range', [-1, 1]))

        h = {v: max_h for v in sampler.nodelist}
        J = {interaction: max_j for interaction in sampler.edgelist}

        return h, J

    def test_sample_ising(self):
        sampler = self.qpu

        h, J = self.nonzero_ising_problem(sampler)

        sampleset = sampler.sample_ising(h, J)

        bqm = dimod.BQM.from_ising(h, J)
        numpy.testing.assert_array_almost_equal(
            bqm.energies(sampleset), sampleset.record.energy)

    def test_sample_qubo(self):
        sampler = self.qpu

        Q, _ = dimod.ising_to_qubo(*self.nonzero_ising_problem(sampler))

        sampleset = sampler.sample_qubo(Q)

        bqm = dimod.BQM.from_qubo(Q)
        numpy.testing.assert_array_almost_equal(
            bqm.energies(sampleset), sampleset.record.energy)

    def test_sample_bqm_ising(self):
        sampler = self.qpu

        bqm = dimod.BQM.from_ising(*self.nonzero_ising_problem(sampler))

        sampleset = sampler.sample(bqm)

        numpy.testing.assert_array_almost_equal(
            bqm.energies(sampleset), sampleset.record.energy)

    def test_sample_bqm_qubo(self):
        sampler = self.qpu

        Q, _ = dimod.ising_to_qubo(*self.nonzero_ising_problem(sampler))
        bqm = dimod.BQM.from_qubo(Q)

        sampleset = sampler.sample(bqm)

        numpy.testing.assert_array_almost_equal(
            bqm.energies(sampleset), sampleset.record.energy)

    def test_mismatched_ising(self):
        sampler = self.qpu

        h = {len(sampler.nodelist)*100: 1}  # get a qubit we know isn't there
        J = {}

        with self.assertRaises(dimod.exceptions.BinaryQuadraticModelStructureError):
            sampler.sample_ising(h, J).resolve()

    def test_mismatched_qubo(self):
        sampler = self.qpu

        v = len(sampler.nodelist)*100
        Q = {(v, v): -1}

        with self.assertRaises(dimod.exceptions.BinaryQuadraticModelStructureError):
            sampler.sample_qubo(Q).resolve()

    def test_problem_labelling(self):
        # NOTE: enable when problem labelling deployed to prod
        if 'cloud' in self.qpu.client.endpoint:
            self.skipTest("labels not supported in production")

        sampler = self.qpu
        h, J = self.nonzero_ising_problem(sampler)
        label = 'problem label'

        # label set
        sampleset = sampler.sample_ising(h, J, label=label)
        self.assertIn('problem_label', sampleset.info)
        self.assertEqual(sampleset.info['problem_label'], label)

        # label unset
        sampleset = sampler.sample_ising(h, J)
        self.assertNotIn('problem_label', sampleset.info)


@unittest.skipIf(os.getenv('SKIP_INT_TESTS'), "Skipping integration test.")
class TestMissingQubits(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        try:
            with Client.from_config() as client:
                solvers = client.get_solvers(qpu=True)

            if not solvers:
                raise unittest.SkipTest("no qpu found")

            # select a QPU with less than 100% yield
            for solver in solvers:
                if solver.num_active_qubits < solver.num_qubits:
                    cls.qpu = DWaveSampler(solver=solver.id)
                    return

            raise unittest.SkipTest("no qpu with less than 100% yield found")

        except (ValueError, ConfigFileError, SolverNotFoundError):
            raise unittest.SkipTest("no qpu available")

    @classmethod
    def tearDownClass(cls):
        cls.qpu.client.close()

    def test_sample_ising_h_list(self):
        sampler = self.qpu

        h = [0 for _ in range(self.qpu.solver.num_qubits)]
        J = {edge: 0 for edge in sampler.edgelist}

        sampleset = sampler.sample_ising(h, J)

        self.assertEqual(set(sampleset.variables), set(sampler.nodelist))
        self.assertLessEqual(len(sampleset.variables), self.qpu.solver.num_qubits)  # sanity check


@unittest.skipIf(os.getenv('SKIP_INT_TESTS'), "Skipping integration test.")
class TestClientSelection(unittest.TestCase):

    def test_client_type(self):
        with mock.patch('dwave.cloud.qpu.Client') as qpu:
            self.assertEqual(DWaveSampler().client, qpu())
            self.assertEqual(DWaveSampler(client='qpu').client, qpu())

        with mock.patch('dwave.cloud.sw.Client') as sw:
            self.assertEqual(DWaveSampler(client='sw').client, sw())

        with mock.patch('dwave.cloud.hybrid.Client') as hybrid:
            self.assertEqual(DWaveSampler(client='hybrid').client, hybrid())

    def test_base_client(self):
        # to test 'base' client instantiation offline,
        # we would need a mock client and a mock solver
        try:
            self.assertEqual(type(DWaveSampler(client=None).client), Client)
            self.assertEqual(type(DWaveSampler(client='base').client), Client)
        except (ValueError, ConfigFileError):
            raise unittest.SkipTest("no API token available")



# Copyright 2020 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import unittest
import unittest.mock as mock
import numpy as np

import dimod
from dwave.cloud import Client
from dwave.cloud.exceptions import SolverNotFoundError

from dwave.system.samplers import LeapHybridSampler
from dwave.system.testing import MockLeapHybridSolver

# Called only for named solver
class MockBadLeapHybridSolver:

    properties = {'category': 'not hybrid'}

class MockClient:

    def __init__(self, **kwargs):

        self.args = kwargs

    def get_solver(self, **filters):

        if self.args.get('solver') == 'not_hybrid_solver':
            return MockBadLeapHybridSolver()

        if self.args.get('client', 'base') not in ['base', 'hybrid']:
            raise SolverNotFoundError

        return MockLeapHybridSolver()

class TestLeapHybridSampler(unittest.TestCase):

    @mock.patch('dwave.system.samplers.leap_hybrid_sampler.Client')
    def test_solver_init(self, mock_client):
        mock_client.from_config.side_effect = MockClient

        default_solver = dict(
            supported_problem_types__contains='bqm',
            order_by='-properties.version')
        self.assertEqual(LeapHybridSampler.default_solver, default_solver)

        defaults = dict(solver=default_solver)

        # Default call
        mock_client.reset_mock()
        LeapHybridSampler()
        mock_client.from_config.assert_called_once_with(
            client='hybrid',
            connection_close=True,
            defaults=defaults)

        # Non-hybrid client setting
        mock_client.reset_mock()
        with self.assertRaises(SolverNotFoundError):
            LeapHybridSampler(client='qpu')

        # Explicitly set solver def
        mock_client.reset_mock()
        LeapHybridSampler(solver={'supported_problem_types__contains': 'bqm'})
        mock_client.from_config.assert_called_once_with(
            client='hybrid',
            solver={'supported_problem_types__contains': 'bqm'},
            connection_close=True,
            defaults=defaults)

        # Named solver
        solver_name = 'hybrid-solver-name'
        mock_client.reset_mock()
        LeapHybridSampler(solver=solver_name)
        mock_client.from_config.assert_called_once_with(
            client='hybrid',
            solver=solver_name,
            connection_close=True,
            defaults=defaults)

        # Named solver: non-hybrid
        with self.assertRaises(ValueError):
            LeapHybridSampler(solver='not_hybrid_solver')

        # Set connection_close to False
        mock_client.reset_mock()
        LeapHybridSampler(connection_close=False)
        mock_client.from_config.assert_called_once_with(
            client='hybrid',
            connection_close=False,
            defaults=defaults)

        mock_client.reset_mock()
        LeapHybridSampler(connection_close=False, solver=solver_name)
        mock_client.from_config.assert_called_once_with(
            client='hybrid',
            solver=solver_name,
            connection_close=False,
            defaults=defaults)

    @mock.patch('dwave.system.samplers.leap_hybrid_sampler.Client')
    def test_sample_bqm(self, mock_client):

        mock_client.from_config.side_effect = MockClient

        bqm = dimod.BinaryQuadraticModel({'a': -1, 'b': 1, 'c': 1},
                    {'ab': -0.8, 'ac': -0.7, 'bc': -1}, 0, dimod.SPIN)

        sampler = LeapHybridSampler()

        response = sampler.sample(bqm)

        rows, cols = response.record.sample.shape

        self.assertEqual(cols, 3)
        self.assertFalse(np.any(response.record.sample == 0))
        self.assertIs(response.vartype, dimod.SPIN)
        self.assertIn('num_occurrences', response.record.dtype.fields)

    @mock.patch('dwave.system.samplers.leap_hybrid_sampler.Client')
    def test_sample_ising_variables(self, mock_client):

        mock_client.from_config.side_effect = MockClient

        sampler = LeapHybridSampler()

        response = sampler.sample_ising({0: -1, 1: 1}, {})

        rows, cols = response.record.sample.shape

        self.assertEqual(cols, 2)

        response = sampler.sample_ising({}, {(0, 4): 1})

        rows, cols = response.record.sample.shape
        self.assertEqual(cols, 2)
        self.assertFalse(np.any(response.record.sample == 0))
        self.assertIs(response.vartype, dimod.SPIN)

    @mock.patch('dwave.system.samplers.leap_hybrid_sampler.Client')
    def test_sample_qubo_variables(self, mock_client):

        mock_client.from_config.side_effect = MockClient

        sampler = LeapHybridSampler()

        response = sampler.sample_qubo({(0, 0): -1, (1, 1): 1})

        rows, cols = response.record.sample.shape

        self.assertEqual(cols, 2)

        response = sampler.sample_qubo({(0, 0): -1, (1, 1): 1})

        rows, cols = response.record.sample.shape

        self.assertEqual(cols, 2)
        self.assertTrue(np.all(response.record.sample >= 0))
        self.assertIs(response.vartype, dimod.BINARY)



# Copyright 2019 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import unittest

import dimod

from dwave.system import PolyCutOffComposite


class CutoffChecker(dimod.PolySampler):
    def __init__(self, child_sampler, expected_poly):

        self.child = child_sampler
        self.poly = expected_poly

    def sample_poly(self, poly, **parameters):
        assert self.poly == poly, '{} != {}'.format(self.poly, poly)
        return self.child.sample_poly(poly, **parameters)

    def parameters(self):
        return self.child.parameters()

    def properties(self):
        return self.child.properties()


class TestConstruction(unittest.TestCase):
    def test_instantiation_smoketest(self):
        sampler = PolyCutOffComposite(dimod.HigherOrderComposite(dimod.ExactSolver()), 0)

        self.assertTrue(hasattr(sampler, 'sample_poly'))
        self.assertTrue(hasattr(sampler, 'sample_hising'))
        self.assertTrue(hasattr(sampler, 'sample_hubo'))

    def test_wrap_bqm(self):
        with self.assertRaises(TypeError):
            PolyCutOffComposite(dimod.ExactSolver(), -1)


class TestSampleHising(unittest.TestCase):
    def setUp(self):
        self.child = dimod.HigherOrderComposite(dimod.ExactSolver())

    def test_empty(self):
        h = {}
        J = {}
        cutoff = 1
        expected = dimod.BinaryPolynomial({}, dimod.SPIN)

        checker = CutoffChecker(self.child, expected)
        samples = PolyCutOffComposite(checker, cutoff).sample_hising(h, J)

        self.assertEqual(samples.record.sample.shape[1], 0)  # no variables

    def test_linear(self):
        # they are all isolated
        h = {'a': -1, 'b': .5}
        J = {}
        cutoff = 1

        # we cannot check in this case because all variables are isolated
        # this results in exactly one variable being sent to ExactSolver and
        # we don't know which one it will be, so we just check the correctness
        # of the output
        samples = PolyCutOffComposite(self.child, cutoff).sample_hising(h, J)

        poly = dimod.BinaryPolynomial.from_hising(h, J)
        for sample, energy in samples.data(['sample', 'energy']):
            self.assertAlmostEqual(energy, poly.energy(sample))

    def test_4_path_isolated_tail(self):
        h = {}
        J = {'ab': -1, 'bc': -.5, 'cd': -.5, 'de': -.5}
        cutoff = .75
        expected = dimod.BinaryPolynomial({'ab': -1}, dimod.SPIN)

        checker = CutoffChecker(self.child, expected)
        samples = PolyCutOffComposite(checker, cutoff).sample_hising(h, J)

        poly = dimod.BinaryPolynomial.from_hising(h, J)
        for sample, energy in samples.data(['sample', 'energy']):
            self.assertAlmostEqual(energy, poly.energy(sample))

    def test_triangle(self):
        h = {'a': -1}
        J = {'abde': -1, 'bc': -.5, 'ca': -.5}
        cutoff = .75
        expected = dimod.BinaryPolynomial({'a': -1, 'abde': -1}, dimod.SPIN)

        checker = CutoffChecker(self.child, expected)
        samples = PolyCutOffComposite(checker, cutoff).sample_hising(h, J)

        poly = dimod.BinaryPolynomial.from_hising(h, J)
        for sample, energy in samples.data(['sample', 'energy']):
            self.assertAlmostEqual(energy, poly.energy(sample))

        # 'c' was isolated, should be 1 when restored with the ground state
        self.assertEqual(samples.first.sample['c'], 1)


class TestSamplePoly(unittest.TestCase):
    def test_isolated(self):
        poly = dimod.BinaryPolynomial({'a': 3, 'abc': 4, 'ac': 0.2}, dimod.SPIN)
        sampler = dimod.HigherOrderComposite(dimod.ExactSolver())
        sampleset = PolyCutOffComposite(sampler, 4.1).sample_poly(poly)



# Copyright 2019 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import unittest

from dwave.system.schedules import ramp

class TestRamp(unittest.TestCase):
    def test_typical(self):
        schedule = ramp(.5, .2, 1)
        self.assertEqual(schedule, [(0, 0), (.4, 0), (.6, 1), (1, 1)])

    def test_width_exception(self):
        with self.assertRaises(ValueError):
            ramp(.1, .2, 1)  # curve would begin at (0, 0)
    
    def test_s_exception(self):
        with self.assertRaises(ValueError):
            ramp(-1, 0, 1)





# Copyright 2018 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import unittest

from contextlib import contextmanager

import dimod
import numpy as np

import dwave.embedding


class TestBrokenChains(unittest.TestCase):
    def test_broken_chains_typical(self):
        S = np.array([[-1, 1, -1, 1],
                      [1, 1, -1, -1],
                      [-1, 1, -1, -1]])
        chains = [[0, 1], [2, 3]]

        broken = dwave.embedding.broken_chains(S, chains)

        np.testing.assert_array_equal([[1, 1], [0, 0], [1, 0]], broken)

    def test_broken_chains_chains_length_0(self):
        S = np.array([[-1, 1, -1, 1],
                      [1, 1, -1, -1],
                      [-1, 1, -1, -1]])
        chains = [[0, 1], [], [2, 3]]

        broken = dwave.embedding.broken_chains(S, chains)

        np.testing.assert_array_equal([[1, 0, 1], [0, 0, 0], [1, 0, 0]], broken)

    def test_broken_chains_single_sample(self):
        S = [-1, 1, 1, 1]
        chains = [[0, 1], [2, 3]]

        broken = dwave.embedding.broken_chains(S, chains)

        np.testing.assert_array_equal([[True, False]], broken)

    def test_matrix(self):
        samples_matrix = np.array([[-1, +1, -1, +1],
                                   [+1, +1, +1, +1],
                                   [-1, -1, +1, -1],
                                   [-1, -1, +1, +1]], dtype='int8')
        chain_list = [(0, 1), (2, 3)]

        broken = dwave.embedding.broken_chains(samples_matrix, chain_list)


class TestChainBreakResolutionAPI():
    chains = [[0, 1], [2, 4], [3]]  # needs to be available to MinimizeEnergy

    def test_api(self):
        cases = {}
        cases['ndarray_binary'] = np.triu(np.ones((5, 5)))
        cases['ndarray_spin'] = 2*np.triu(np.ones((5, 5))) - 1
        cases['listoflists_binary'] = [[1, 1, 1, 1, 1],
                                       [0, 1, 1, 1, 1],
                                       [0, 0, 1, 1, 1],
                                       [0, 0, 0, 1, 1],
                                       [0, 0, 0, 0, 1]]

        ss0 = dimod.SampleSet.from_samples(np.triu(np.ones((5, 5))),
                                           energy=0,
                                           vartype=dimod.BINARY)
        cases['sampleset_ordered'] = ss0

        ss1 = dimod.SampleSet.from_samples((np.triu(np.ones((5, 5))),
                                            [0, 4, 1, 3, 2]),
                                           energy=0,
                                           vartype=dimod.BINARY)
        cases['sampleset_unordered'] = ss1

        for description, samples in cases.items():
            with self.subTest(description):
                resolved, idxs = self.chain_break_method(samples, self.chains)

                # correct number of variables (1 per chain)
                self.assertEqual(resolved.shape[1], len(self.chains))

                # correct number of rows
                self.assertEqual(resolved.shape[0], len(idxs))


class TestDiscard(TestChainBreakResolutionAPI, unittest.TestCase, ):

    # for the API tests
    def setUp(self):
        self.chain_break_method = dwave.embedding.discard

    def test_discard_no_breaks_all_ones_identity_embedding(self):

        samples_matrix = np.array(np.ones((100, 50)), dtype='int8')
        chain_list = [[idx] for idx in range(50)]

        new_matrix, idxs = dwave.embedding.discard(samples_matrix, chain_list)

        np.testing.assert_equal(new_matrix, samples_matrix)

    def test_discard_no_breaks_all_ones_one_var_embedding(self):

        samples_matrix = np.array(np.ones((100, 50)), dtype='int8')
        chain_list = [[idx for idx in range(50)]]

        new_matrix, idxs = dwave.embedding.discard(samples_matrix, chain_list)

        self.assertEqual(new_matrix.shape, (100, 1))

    def test_discard_typical(self):

        samples_matrix = np.array([[-1, +1, -1, +1],
                                   [+1, +1, +1, +1],
                                   [-1, -1, +1, -1],
                                   [-1, -1, +1, +1]], dtype='int8')
        chain_list = [(0, 1), (2, 3)]

        new_matrix, idxs = dwave.embedding.discard(samples_matrix, chain_list)

        np.testing.assert_equal(new_matrix, [[+1, +1],
                                             [-1, +1]])

    def test_mixed_chain_types(self):
        chains = [(0, 1), [2, 3], {4, 5}]
        samples = [[1, 1, 1, 1, 1, 1], [0, 1, 0, 1, 0, 1]]
        unembedded, idx = dwave.embedding.discard(samples, chains)

        np.testing.assert_array_equal(unembedded, [[1, 1, 1]])
        np.testing.assert_array_equal(idx, [0])


class TestMajorityVote(TestChainBreakResolutionAPI, unittest.TestCase):

    # for the API tests
    def setUp(self):
        self.chain_break_method = dwave.embedding.majority_vote

    def test_typical_spin(self):
        S = np.array([[-1, +1, -1, +1],
                      [+1, +1, -1, +1],
                      [-1, +1, -1, -1]])
        chains = [[0, 1, 2], [3]]

        samples, idx = dwave.embedding.majority_vote(S, chains)

        np.testing.assert_equal(samples, [[-1, +1],
                                          [+1, +1],
                                          [-1, -1]])

    def test_typical_binary(self):
        S = np.array([[0, 1, 0, 1],
                      [1, 1, 0, 1],
                      [0, 1, 0, 0]])
        chains = [[0, 1, 2], [3]]

        samples, idx = dwave.embedding.majority_vote(S, chains)

        np.testing.assert_equal(samples, [[0, 1],
                                          [1, 1],
                                          [0, 0]])

    def test_four_chains(self):
        S = [[-1, -1, -1, -1],
             [+1, -1, -1, -1],
             [+1, +1, -1, -1],
             [-1, +1, -1, -1],
             [-1, +1, +1, -1],
             [+1, +1, +1, -1],
             [+1, -1, +1, -1],
             [-1, -1, +1, -1],
             [-1, -1, +1, +1],
             [+1, -1, +1, +1],
             [+1, +1, +1, +1],
             [-1, +1, +1, +1],
             [-1, +1, -1, +1],
             [+1, +1, -1, +1],
             [+1, -1, -1, +1],
             [-1, -1, -1, +1]]
        chains = [[0], [1], [2, 3]]

        samples, idx = dwave.embedding.majority_vote(S, chains)

        self.assertEqual(samples.shape, (16, 3))
        self.assertEqual(set().union(*samples), {-1, 1})  # should be spin-valued


class TestMinimizeEnergy(TestChainBreakResolutionAPI, unittest.TestCase):
    # for the API tests
    def setUp(self):
        embedding = dict(zip('abcdefghijk', self.chains))
        bqm = dimod.BinaryQuadraticModel.from_ising({v: 0 for v in embedding}, {})
        self.chain_break_method = dwave.embedding.MinimizeEnergy(bqm, embedding)

    def test_minimize_energy(self):
        embedding = {0: (0, 5), 1: (1, 6), 2: (2, 7), 3: (3, 8), 4: (4, 10)}
        h = []
        j = {(0, 1): -1, (0, 2): 2, (0, 3): 2, (0, 4): -1,
             (2, 1): -1, (1, 3): 2, (3, 1): -1, (1, 4): -1,
             (2, 3): 1, (4, 2): -1, (2, 4): -1, (3, 4): 1}

        bqm = dimod.BinaryQuadraticModel.from_ising(h, j)

        solutions = [
            [-1, -1, -1, -1, -1, -1, +1, +1, +1, 3, +1],
            [+1, +1, +1, +1, +1, -1, +1, -1, -1, 3, -1],
            [+1, +1, -1, +1, -1, -1, -1, -1, -1, 3, -1]
        ]
        expected = [
            [-1, -1, +1, +1, -1],
            [+1, +1, +1, -1, +1],
            [-1, -1, -1, +1, -1]
        ]

        cbm = dwave.embedding.MinimizeEnergy(bqm, embedding)

        unembedded, idx = cbm(solutions, [embedding[v] for v in range(5)])

        np.testing.assert_array_equal(expected, unembedded)

    def test_minimize_energy_non_clique(self):
        embedding = {0: (0, 5), 1: (1, 6), 2: (2, 7), 3: (3, 8), 4: (4, 10)}
        h = []
        j = {(0, 1): -1, (0, 2): 2, (0, 3): 2, (0, 4): -1,
             (1, 3): 2, (3, 1): -1, (1, 4): -1,
             (2, 3): 1, (4, 2): -1, (2, 4): -1, (3, 4): 1}

        bqm = dimod.BinaryQuadraticModel.from_ising(h, j)

        solutions = [
            [-1, -1, -1, -1, -1, -1, +1, +1, +1, 3, +1],
            [+1, +1, +1, +1, +1, -1, +1, -1, -1, 3, -1],
            [+1, +1, -1, +1, -1, -1, -1, -1, -1, 3, -1]
        ]

        expected = [
            [-1, -1, +1, +1, -1],
            [+1, +1, +1, -1, +1],
            [-1, -1, -1, +1, -1]
        ]

        cbm = dwave.embedding.MinimizeEnergy(bqm, embedding)
        unembedded, idx = cbm(solutions, [embedding[v] for v in range(5)])

        np.testing.assert_array_equal(expected, unembedded)

    def test_minimize_energy_easy(self):
        chains = ({0, 1}, [2], (4, 5, 6))
        embedding = {v: chain for v, chain in enumerate(chains)}
        h = [-1, 0, 0]
        j = {}
        bqm = dimod.BinaryQuadraticModel.from_ising(h, j)
        solutions = [
            [-1, -1, +1, 3, -1, -1, -1],
            [-1, +1, -1, 3, +1, +1, +1]
        ]
        expected = [
            [-1, +1, -1],
            [+1, -1, +1]
        ]
        cbm = dwave.embedding.MinimizeEnergy(bqm, embedding)

        unembedded, idx = cbm(solutions, chains)

        np.testing.assert_array_equal(expected, unembedded)

    def test_empty_matrix(self):
        chains = []
        bqm = dimod.BinaryQuadraticModel.empty(dimod.BINARY)
        solutions = [[]]
        embedding = {}
        cbm = dwave.embedding.MinimizeEnergy(bqm, embedding)
        unembedded, idx = cbm(solutions, chains)

        np.testing.assert_array_equal([[]], unembedded)
        np.testing.assert_array_equal(idx, [0])

    def test_empty_chains(self):
        embedding = {}
        h = []
        j = {(0, 1): -1, (0, 2): 2, (0, 3): 2, (0, 4): -1,
             (2, 1): -1, (1, 3): 2, (3, 1): -1, (1, 4): -1,
             (2, 3): 1, (4, 2): -1, (2, 4): -1, (3, 4): 1}

        bqm = dimod.BinaryQuadraticModel.from_ising(h, j)

        solutions = [
            [-1, -1, -1, -1, -1, -1, +1, +1, +1, 3, +1],
            [+1, +1, +1, +1, +1, -1, +1, -1, -1, 3, -1],
            [+1, +1, -1, +1, -1, -1, -1, -1, -1, 3, -1]
        ]
        expected = [
            [-1, -1, +1, +1, -1],
            [+1, +1, +1, -1, +1],
            [-1, -1, -1, +1, -1]
        ]

        cbm = dwave.embedding.MinimizeEnergy(bqm, embedding)

        unembedded, idx = cbm(solutions, [])

        np.testing.assert_array_equal(unembedded, [[], [], []])

    def test_unordered_labels(self):
        Q = {(0, 1): 1, (0, 2): 1, (0, 3): 1, (1, 2): 1, (1, 3): 1, (2, 3): 1,
             (0, 0): 1, (1, 1): 1, (2, 2): 1, (3, 3): 1}
        bqm = dimod.BinaryQuadraticModel.from_qubo(Q)

        embedding = {0: [55], 1: [48], 2: [50, 53], 3: [52, 51]}

        samples = (np.triu(np.ones(6)), [48, 50, 51, 52, 53, 55])
        sampleset = dimod.SampleSet.from_samples(samples, energy=0,
                                                 vartype=dimod.BINARY)

        cbm = dwave.embedding.MinimizeEnergy(bqm, embedding)

        unembedded, idx = cbm(sampleset, [[55], [48], [50, 53], [52, 51]])


class TestWeightedRandom(TestChainBreakResolutionAPI, unittest.TestCase):
    # for the API tests
    def setUp(self):
        self.chain_break_method = dwave.embedding.weighted_random



# Copyright 2019 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import enum
import logging

import dimod
import numpy as np
import collections.abc as abc

from dwave.embedding import broken_chains


class WarningAction(enum.Enum):
    """Settings for raising warnings.

    An enum with values ``IGNORE`` and ``SAVE``.
    """

    IGNORE = 'ignore'
    SAVE = 'save'

    # we may eventually want to support logging and raising python Warnings
    # LOG = 'log'
    # RAISE = 'raise'


IGNORE = WarningAction.IGNORE
SAVE = WarningAction.SAVE


# LOG = WarningAction.LOG
# RAISE = WarningAction.RAISE


def as_action(action):
    if isinstance(action, WarningAction):
        return action
    elif isinstance(action, str):
        return WarningAction[action.upper()]
    else:
        raise TypeError('unknown warning action provided')


class ChainBreakWarning(UserWarning):
    """Raised if a chain's qubits are in different states for lowest-energy samples."""
    pass


class ChainLengthWarning(UserWarning):
    """Raised if the number of qubits forming a chain is high."""
    pass


class TooFewSamplesWarning(UserWarning):
    """Raised if lowest-energy samples are a small fraction of the total samples."""
    pass


class ChainStrengthWarning(UserWarning):
    """Base category for warnings about the embedding chain strength."""
    pass


class EnergyScaleWarning(UserWarning):
    """Base category for warnings about the relative bias strengths."""
    pass


class WarningHandler(object):
    def __init__(self, action=None):
        self.saved = []

        if action is not None:
            # promote from class attribute to object attribute
            self.action = as_action(action)

    action = WarningAction.IGNORE  # the default

    # todo: let user override __init__ parameters with kwargs
    def issue(self, msg, category=None, func=None, level=logging.WARNING,
              data=None):
        """Issue a warning.

        Args:
            msg (str):
                The warning message

            category (Warning):
                The warning category class. Defaults to UserWarning.

            level (int):
                The level of warning severity. Uses the logging warning levels.

            func (function):
                A function that is executed in the case that the warning level
                is not IGNORE. The function should return a 2-tuple containing
                a bool specifying whether the warning should be saved/raised
                and any relevent data associated with the warning as a
                dictionary/None. This overrides anything provided in the `data`
                kwarg.

            data (dict):
                Any data relevent to the warning.

        """

        action = as_action(self.action)  # user may have overwritten

        if action is IGNORE:
            return

        if func is not None:
            valid, data = func()
            if not valid:
                return

        if category is None:
            category = UserWarning

        if data is None:
            data = {}

        if action is SAVE:
            self.saved.append(dict(type=category,
                                   message=msg,
                                   level=level,
                                   data=data))
        else:
            raise TypeError("unknown action")

    # some hard-coded warnings for convenience or for expensive operations

    def chain_length(self, embedding, length=7):
        if as_action(self.action) is IGNORE:
            return

        for v, chain in embedding.items():
            if len(chain) <= length:
                continue

            self.issue("Chain length greater than {}".format(length),
                       category=ChainLengthWarning,
                       data=dict(target_variables=chain,
                                 source_variables=[v]),
                       )

    def chain_break(self, sampleset, embedding):
        if as_action(self.action) is IGNORE:
            return

        ground = sampleset.lowest()
        variables = list(embedding)
        chains = [embedding[v] for v in variables]
        broken = broken_chains(ground, chains)

        if not (len(sampleset) and broken.any()):
            return

        for nc, chain in enumerate(chains):
            for row in range(broken.shape[0]):
                if not broken[row, nc]:
                    continue

                self.issue("Lowest-energy samples contain a broken chain",
                           category=ChainBreakWarning,
                           level=logging.ERROR,
                           data=dict(target_variables=chain,
                                     source_variables=[variables[nc]],
                                     sample_index=row),
                           )

    def chain_strength(self, bqm, chain_strength, embedding=None):
        """Issues a warning when any quadratic biases are greater than the given
        chain strength."""
        if as_action(self.action) is IGNORE:
            return

        if embedding is not None:
            if not embedding or all(len(chain) <= 1 for chain in embedding.values()):
                # the chains are all length 1 so don't have to worry about
                # strength
                return

        if isinstance(chain_strength, abc.Mapping):
            interactions = [(u, v) for (u, v), bias in bqm.quadratic.items()
                            if abs(bias) >= min(chain_strength[u], chain_strength[v])]
        else:
            interactions = [uv for uv, bias in bqm.quadratic.items()
                            if abs(bias) >= chain_strength]

        if interactions:
            self.issue("Some quadratic biases are stronger than the given "
                       "chain strength",
                       category=ChainStrengthWarning,
                       level=logging.WARNING,
                       data=dict(source_interactions=interactions))

    def energy_scale(self, bqm):
        """Issues a warning if some biases are 10^3 times stronger than others.

        Args:
            bqm (:class:`dimod.BinaryQuadraticModel`/tuple):
                A binary quadratic model, a tuple of the form `(Q)` where `Q`
                is a QUBO-dictionary, or a tuple of the form `(h, J)` where
                `h` and `J` are Ising problem dictionaries.

        """
        if as_action(self.action) is IGNORE:
            return

        if isinstance(bqm, tuple):
            if len(bqm) == 1:
                bqm = dimod.BinaryQuadraticModel.from_qubo(*bqm)
            elif len(bqm) == 2:
                bqm = dimod.BinaryQuadraticModel.from_ising(*bqm)
            else:
                raise TypeError("bqm should be a binary quadratic model, a "
                                "1-tuple or a 2-tuple")

        max_bias = max(map(abs, bqm.linear.values()))
        if bqm.quadratic:
            max_bias = max(max_bias, max(map(abs, bqm.quadratic.values())))

        max_bias *= 10 ** -3

        variables = [v for v, bias in bqm.linear.items()
                     if abs(bias) < max_bias]
        interactions = [uv for uv, bias in bqm.quadratic.items()
                        if abs(bias) < max_bias]

        data = dict()
        if variables:
            data.update(source_variables=variables)
        if interactions:
            data.update(source_interactions=interactions)

        if data:
            self.issue("Some biases are 10^3 times stronger than others",
                       category=EnergyScaleWarning,
                       level=logging.WARNING,
                       data=data)

    def too_few_samples(self, sampleset):
        """Issues a warning when the number ground states found is within the sampling error threshold."""
        if self.action is IGNORE:
            return

        ground = sampleset.lowest()
        total_ground = np.sum(ground.record.num_occurrences)
        total_samples = np.sum(sampleset.record.num_occurrences)

        if total_ground <= np.sqrt(total_samples):
            self.issue("Number of ground states found is within sampling error",
                       category=TooFewSamplesWarning,
                       level=logging.WARNING,
                       data=dict(number_of_ground_states=total_ground,
                                 num_reads=total_samples,
                                 sampling_error_rate=np.sqrt(total_samples)),
                       )



# Copyright 2018 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import unittest
import collections

import dimod

from dwave.system.composites import VirtualGraphComposite

from dwave.system.testing import MockDWaveSampler


class TestVirtualGraphWithMockDWaveSampler(unittest.TestCase):
    def setUp(self):
        self.sampler = MockDWaveSampler()

    def test_smoke(self):
        child_sampler = MockDWaveSampler()
        sampler = VirtualGraphComposite(child_sampler, {'a': [0]}, flux_bias_num_reads=1)

        # depending on how recenlty flux bias data was gathered, this may be true
        child_sampler.flux_biases_flag = False

        if sampler.flux_biases:
            sampler.sample_ising({'a': -1}, {})
            self.assertTrue(child_sampler.flux_biases_flag)  # true when some have been provided to sample_ising

    def test_structure_keyword_setting(self):
        sampler = VirtualGraphComposite(self.sampler, embedding={'a': set(range(8)),
                                                                 'b': set(range(8, 16)),
                                                                 'c': set(range(16, 24))},
                                        flux_biases=False)

        nodelist, edgelist, adj = sampler.structure
        self.assertEqual(nodelist, ['a', 'b', 'c'])
        self.assertEqual(edgelist, [('a', 'b'), ('b', 'c')])
        self.assertEqual(adj, {'a': {'b'}, 'b': {'a', 'c'}, 'c': {'b'}})

        # unlike variable names
        sampler = VirtualGraphComposite(self.sampler, embedding={'a': set(range(8)),
                                                                 1: set(range(8, 16)),
                                                                 'c': set(range(16, 24))},
                                        flux_biases=False)
        nodelist, edgelist, adj = sampler.structure
        self.assertEqual(set(nodelist), {'a', 1, 'c'})
        self.assertEqual(adj, {'a': {1}, 1: {'a', 'c'}, 'c': {1}})
        self.assertIsInstance(edgelist, list)
        self.assertIsInstance(nodelist, list)

        # edges should still be unique
        for u in adj:
            for v in adj[u]:
                if (u, v) in edgelist:
                    assert (v, u) not in edgelist
                if (v, u) in edgelist:
                    assert (u, v) not in edgelist
                assert (u, v) in edgelist or (v, u) in edgelist

    def test_embedding_parameter(self):
        """Given embedding should be saved as a parameter"""
        sampler = self.sampler

        __, __, adj = sampler.structure
        embedding = {v: (v,) for v in adj}

        sampler = VirtualGraphComposite(sampler, embedding=embedding, flux_biases=False)

        self.assertEqual(sampler.embedding, embedding)

    def test_simple_complete_graph_sample_ising(self):
        """sample_ising on a K4."""

        K4 = VirtualGraphComposite(self.sampler, embedding={0: {0, 4},
                                                            1: {1, 5},
                                                            2: {2, 6},
                                                            3: {3, 7}},
                                   flux_biases=False)

        K4.sample_ising({0: .1, 1: .2}, {(0, 1): 1.5})


class Test_ValidateChainStrength(unittest.TestCase):

    def test_no_properties(self):
        from dwave.system.composites.virtual_graph import _validate_chain_strength

        Sampler = collections.namedtuple('Sampler', ['properties'])
        sampler = Sampler({})

        with self.assertRaises(ValueError):
            _validate_chain_strength(sampler, None)

        with self.assertRaises(ValueError):
            _validate_chain_strength(sampler, 1.0)

    def test_j_range(self):
        from dwave.system.composites.virtual_graph import _validate_chain_strength

        Sampler = collections.namedtuple('Sampler', ['properties'])
        sampler = Sampler({'j_range': [-1.0, 1.0]})

        self.assertEqual(_validate_chain_strength(sampler, None), 1.0)
        self.assertEqual(_validate_chain_strength(sampler, .5), .5)

        with self.assertRaises(ValueError):
            _validate_chain_strength(sampler, 1.5)

    def test_extended_j_range(self):
        from dwave.system.composites.virtual_graph import _validate_chain_strength

        Sampler = collections.namedtuple('Sampler', ['properties'])
        sampler = Sampler({'j_range': [-1.0, 1.0], 'extended_j_range': [-2.0, 2.0]})

        self.assertEqual(_validate_chain_strength(sampler, None), 2.0)
        self.assertEqual(_validate_chain_strength(sampler, .5), .5)
        self.assertEqual(_validate_chain_strength(sampler, 1.5), 1.5)



# Copyright 2019 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""Utility functions."""

import os
import json
import networkx as nx

__all__ = ['common_working_graph', 'classproperty']


# taken from https://stackoverflow.com/a/39542816, licensed under CC BY-SA 3.0
# not needed in py39+
class classproperty(property):
    def __get__(self, obj, objtype=None):
        return super(classproperty, self).__get__(objtype)


def common_working_graph(graph0, graph1):
    """Creates a graph using the common nodes and edges of two given graphs.

    This function finds the edges and nodes with common labels. Note that this
    not the same as finding the greatest common subgraph with isomorphisms.

    Args:
        graph0: (dict[dict]/:obj:`~networkx.Graph`)
            A NetworkX graph or a dictionary of dictionaries adjacency
            representation.

        graph1: (dict[dict]/:obj:`~networkx.Graph`)
            A NetworkX graph or a dictionary of dictionaries adjacency
            representation.

    Returns:
        :obj:`~networkx.Graph`: A graph with the nodes and edges common to both
        input graphs.

    Examples:

        This example creates a graph that represents a quarter (4 by 4 Chimera tiles)
        of a particular D-Wave system's working graph.

        >>> import dwave_networkx as dnx
        >>> from dwave.system import DWaveSampler, common_working_graph
        ...
        >>> sampler = DWaveSampler()
        >>> C4 = dnx.chimera_graph(4)  # a 4x4 lattice of Chimera tiles
        >>> c4_working_graph = common_working_graph(C4, sampler.adjacency)   

    """

    G = nx.Graph()
    G.add_nodes_from(v for v in graph0 if v in graph1)
    G.add_edges_from((u, v) for u in graph0 for v in graph0[u]
                     if v in graph1 and u in graph1[v])

    return(G)


class FeatureFlags:
    """User environment-level Ocean feature flags pertinent to dwave-system."""

    # NOTE: This is an experimental feature. If we decide to keep it, we'll want
    # to move this machinery level up to Ocean-common.

    @staticmethod
    def get(name, default=False):
        try:
            return json.loads(os.getenv('DWAVE_FEATURE_FLAGS')).get(name, default)
        except:
            return default

    @classproperty
    def hss_solver_config_override(cls):
        return cls.get('hss_solver_config_override')



# Copyright 2018 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import os

import homebase

from dwave.system.package_info import __version__

__all__ = ['cache_file']

APPNAME = 'dwave-virtual-graph-cache'
"""The application name is used to determine the cache location."""

APPAUTHOR = 'dwave-systems'
"""The application author is used to determine the cache location."""

DATABASENAME = 'virtual-graph_v%s.db' % __version__
"""The name for the sqlite database itself. Based on the version of the package."""


def cache_file(app_name=APPNAME, app_author=APPAUTHOR, filename=DATABASENAME):
    """Returns the filename (including path) for the data cache.

    The path will depend on the operating system, certain environmental
    variables and whether it is being run inside a virtual environment.
    See `homebase <https://github.com/dwavesystems/homebase>`_.

    Args:
        app_name (str, optional): The application name.
            Default is given by :obj:`.APPNAME`.
        app_author (str, optional): The application author. Default
            is given by :obj:`.APPAUTHOR`.
        filename (str, optional): The name of the database file.
            Default is given by :obj:`DATABASENAME`.

    Returns:
        str: The full path to the file that can be used as a cache.

    Notes:
        Creates the directory if it does not already exist.

        If run inside of a virtual environment, the cache will be stored
        in `/path/to/virtualenv/data/app_name`

    """
    user_data_dir = homebase.user_data_dir(app_name=app_name, app_author=app_author, create=True)
    return os.path.join(user_data_dir, filename)
